"""``aguaclara`` package playground setup

Provide easy setup for using the ``aguaclara`` package and all of its modules,
classes, and functions.

Note: this module should not be used for development. Instead, individually
import only necessary modules.

Example:
    >>> from aguaclara.play import *
"""

import math
import numpy as np
import pandas as pd
import matplotlib
import matplotlib.style
import matplotlib.pyplot as plt
import warnings

import aguaclara.core.constants as con
from aguaclara.core import drills
import aguaclara.core.head_loss as k
import aguaclara.core.materials as mat
from aguaclara.core import physchem as pc
from aguaclara.core import pipeline
import aguaclara.core.pipes as pipe
from aguaclara.core.units import unit_registry as u
import aguaclara.core.utility as ut

# Temporarily disabled for release 0.0.15 to prevent problems with lacking
# Onshapepy setup. Re-enable when Onshapepy backend has been resolved.
# from aguaclara.design.lfom import LFOM
# from aguaclara.design.floc import Flocculator
import aguaclara.design.human_access as ha

import aguaclara.research.environmental_processes_analysis as epa
import aguaclara.research.floc_model as fm
import aguaclara.research.procoda_parser as procoda_parser
import aguaclara.research.peristaltic_pump as peristaltic_pump
import aguaclara.research.stock_qc as stock_qc

def set_sig_figs(n=4):
    """Set the number of significant figures used to print Pint, Pandas, and
    NumPy quantities.

    Args:
        n (int): Number of significant figures to display.
    """
    u.default_format = '.' + str(n) + 'g'
    pd.options.display.float_format = ('{:,.' + str(n) + '}').format


set_sig_figs()
matplotlib.style.use('ggplot')



'''''
research tests.
'''''

import unittest
from aguaclara.research.environmental_processes_analysis import *


class TestEPA(unittest.TestCase):
    '''''
    Test research's Environmental_Processes_Analysis
    '''''

    def test_Hplus_concentration_from_pH(self):
        '''''
        Test function that converts pH to molarity of H+
        '''''
        answer = invpH(8.25)
        self.assertEqual(answer, 5.623413251903491e-09*u.mol/u.L)

        answer = invpH(10)
        self.assertEqual(answer, 1e-10*u.mol/u.L)



"""This file provides basic utility functions such as significant figures which
can be used throughout the plant design.

"""
from aguaclara.core.units import unit_registry as u


import numpy as np
from math import log10, floor


def round_sf(number, digits):
    """Returns inputted value rounded to number of significant figures desired.

    :param number: Value to be rounded
    :type number: float
    :param digits: number of significant digits to be rounded to.
    :type digits: int
    """
    units = None
    try:
        num = number.magnitude
        units = number.units
    except AttributeError:
        num = number

    try:
        if (units != None):
            rounded_num = round(num, digits - int(floor(log10(abs(num)))) - 1) * units
        else:
            rounded_num = round(num, digits - int(floor(log10(abs(num)))) - 1)
        return rounded_num
    except ValueError:  # Prevents an error with log10(0)
        if (units != None):
            return 0 * units
        else:
            return 0


def stepceil_with_units(param, step, unit):
    """This function returns the smallest multiple of 'step' greater than or
    equal to 'param' and outputs the result in Pint units.
    This function is unit-aware and functions without requiring translation
    so long as 'param' and 'unit' are of the same dimensionality.
    """
    counter = 0 * unit
    while counter < param.to(unit):
        counter += step * unit
    return counter


# Take the values of the array, compare to x, find the index of the first value less than or equal to x
def floor_nearest(x,array):
    myindex = np.argmax(array >= x) - 1
    return array[myindex]


# Take the values of the array, compare to x, find the index of the first value greater or equal to x
def ceil_nearest(x,array):
    myindex = np.argmax(array >= x)
    return array[myindex]


def list_handler(HandlerResult="nparray"):
    """Wraps a function to handle list inputs."""
    def decorate(func):
        def wrapper(*args, **kwargs):
            """Run through the wrapped function once for each array element.

            :param HandlerResult: output type. Defaults to numpy arrays.
            """
            sequences = []
            enumsUnitCheck = enumerate(args)
            argsList = list(args)
            #This for loop identifies pint unit objects and strips them
            #of their units.
            for num, arg in enumsUnitCheck:
                if type(arg) == type(1 * u.m):
                    argsList[num] = arg.to_base_units().magnitude
            enumsUnitless = enumerate(argsList)
            #This for loop identifies arguments that are sequences and
            #adds their index location to the list 'sequences'.
            for num, arg in enumsUnitless:
                if isinstance(arg, (list, tuple, np.ndarray)):
                    sequences.append(num)
            #If there are no sequences to iterate through, simply return
            #the function.
            if len(sequences) == 0:
                result = func(*args, **kwargs)
            else:
                #iterant keeps track of how many times we've iterated and
                #limiter stops the loop once we've iterated as many times
                #as there are list elements. Without this check, a few
                #erroneous runs will occur, appending the last couple values
                #to the end of the list multiple times.
                #
                #We only care about the length of sequences[0] because this
                #function is recursive, and sequences[0] is always the relevant
                #sequences for any given run.
                limiter = len(argsList[sequences[0]])
                iterant = 0
                result = []
                for num in sequences:
                    for arg in argsList[num]:
                        if iterant >= limiter:
                            break
                        #We can safely replace the entire list argument
                        #with a single element from it because of the looping
                        #we're doing. We redefine the object, but that
                        #definition remains within this namespace and does
                        #not penetrate further up the function.
                        argsList[num] = arg
                        #Here we dive down the rabbit hole. This ends up
                        #creating a multi-dimensional array shaped by the
                        #sizes and shapes of the lists passed.
                        result.append(wrapper(*argsList,
                                              HandlerResult=HandlerResult, **kwargs))
                        iterant += 1
                #HandlerResult allows the user to specify what type to
                #return the generated sequence as. It defaults to numpy
                #arrays because functions tend to handle them better, but if
                #the user does not wish to import numpy the base Python options
                #are available to them.
                if HandlerResult == "nparray":
                    result = np.array(result)
                elif HandlerResult == "tuple":
                    result = tuple(result)
                elif HandlerResult == "list":
                    result == list(result)
            return result
        return wrapper
    return decorate


def check_range(*args):
    """
    Check whether passed paramters fall within approved ranges.

    Does not return anything, but will raise an error if a parameter falls
    outside of its defined range.

    Input should be passed as an array of sequences, with each sequence
    having three elements:
        [0] is the value being checked,
        [1] is the range parameter(s) within which the value should fall, and
        [2] is the name of the parameter, for better error messages.
    If [2] is not supplied, "Input" will be appended as a generic name.

    Range requests that this function understands are listed in the
    knownChecks sequence.
    """
    knownChecks = ('>0', '>=0', '0-1', '<0', '<=0', 'int', 'boolean')
    for arg in args:
        #Converts arg to a mutable list
        arg = [*arg]
        if len(arg) == 1:
            #arg[1] details what range the parameter should fall within; if
            #len(arg) is 1 that means a validity was not specified and the
            #parameter should not have been passed in its current form
            raise TypeError("No range-validity parameter provided.")
        elif len(arg) == 2:
            #Appending 'Input" to the end allows us to give more descriptive
            #error messages that do not fail if no description was supplied.
            arg.append("Input")
        #This ensures that all whitespace is removed before checking if the
        #request is understood
        arg[1] = "".join(arg[1].lower().split())
        #This block checks that each range request is understood.
        #If the request is a compound one, it must be separated into individual
        #requests for validity comprehension
        for i in arg[1].split(","):
            if i not in knownChecks:
                raise RuntimeError("Unknown parameter validation "
                                       "request: {0}.".format(i))
        if not isinstance(arg[0], (list, tuple, np.ndarray)):
            arg[0] = [arg[0]]
        for i in arg[0]:
            if '>0' in arg[1] and i <= 0:
                raise ValueError("{1} is {0} but must be greater than "
                                 "0.".format(i, arg[2]))
            if '>=0' in arg[1] and i <0:
                raise ValueError("{1} is {0} but must be 0 or "
                                 "greater.".format(i, arg[2]))
            if '0-1' in arg[1] and not 0 <= i <= 1:
                raise ValueError("{1} is {0} but must be between 0 and "
                                 "1.".format(i, arg[2]))
            if '<0' in arg[1] and i >= 0:
                raise ValueError("{1} is {0} but must be less than "
                                 "0.".format(i, arg[2]))
            if '<=0' in arg[1] and i >0:
                raise ValueError("{1} is {0} but must be 0 or "
                                 "less.".format(i, arg[2]))
            if 'int' in arg[1] and int(i) != i:
                raise TypeError("{1} is {0} but must be a numeric "
                                "integer.".format(i, arg[2]))
            if 'boolean' in arg[1] and type(i) != bool:
                raise TypeError("{1} is {0} but must be a "
                                "boolean.".format(i, arg[2]))



"""This file contains all the functions needed to design an entrance tank for
an AguaClara plant.

"""
import design.ent_tank
from aguaclara.play import*

@u.wraps(u.inch, [u.m**3/u.s, u.degK, u.m, None], False)
def drain_OD(q_plant, T, depth_end, SDR):
    """Return the nominal diameter of the entrance tank drain pipe. Depth at the
    end of the flocculator is used for headloss and length calculation inputs in
    the diam_pipe calculation.

    Parameters
    ----------
    q_plant: float
        Plant flow rate

    T: float
        Design temperature

    depth_end: float
        The depth of water at the end of the flocculator

    SDR: float
        Standard dimension ratio

    Returns
    -------
    float
        ?

    Examples
    --------
    >>> from aguaclara.play import*
    ??
    """
    nu = pc.viscosity_kinematic(T)
    K_minor = con.PIPE_ENTRANCE_K_MINOR + con.PIPE_EXIT_K_MINOR + con.EL90_K_MINOR
    drain_ID = pc.diam_pipe(q_plant, depth_end, depth_end, nu, mat.PVC_PIPE_ROUGH, K_minor)
    drain_ND = pipe.SDR_available_ND(drain_ID, SDR)
    return pipe.OD(drain_ND).magnitude

@u.wraps(None, [u.m**3/u.s, u.m], False)
def num_plates_ET(q_plant, W_chan):
    """Return the number of plates in the entrance tank.

    This number minimizes the total length of the plate settler unit.

    Parameters
    ----------
    q_plant: float
        Plant flow rate

    W_chan: float
        Width of channel

    Returns
    -------
    float
        ?

    Examples
    --------
    >>> from aguaclara.play import*
    >>> num_plates_ET(20*u.L/u.s,2*u.m)
    1.0
    """
    num_plates = np.ceil(np.sqrt(q_plant / (design.ent_tank.CENTER_PLATE_DIST.magnitude
                                            * W_chan * design.ent_tank.CAPTURE_BOD_VEL.magnitude * np.sin(
                design.ent_tank.PLATE_ANGLE.to(u.rad).magnitude))))
    return num_plates

@u.wraps(u.m, [u.m**3/u.s, u.m], False)
def L_plate_ET(q_plant, W_chan):
    """Return the length of the plates in the entrance tank.

    Parameters
    ----------
    q_plant: float
        Plant flow rate

    W_chan: float
        Width of channel

    Returns
    -------
    float
        ?

    Examples
    --------
    >>> from aguaclara.play import*
    >>> L_plate_ET(20*u.L/u.s,2*u.m)
    0.00194
    """
    L_plate = (q_plant / (num_plates_ET(q_plant, W_chan) * W_chan *
                          design.ent_tank.CAPTURE_BOD_VEL.magnitude * np.cos(
                design.ent_tank.PLATE_ANGLE.to(u.rad).magnitude)))
    - (design.ent_tank.PLATE_S.magnitude * np.tan(design.ent_tank.PLATE_ANGLE.to(u.rad).magnitude))
    return L_plate



"""Flocculator design.

This module provides all constants and functions necessary to design an
AguaClara flocculator. These constants and functions define both hydraulic
(head loss, retention time, etc.) and geometric (baffle spacing, channel length,
etc.) values that specify a flocculator design.
"""

import aguaclara.core.head_loss as minorloss
import aguaclara.design.human_access as ha
import aguaclara.core.physchem as pc
import aguaclara.core.pipes as pipes
from aguaclara.core.units import unit_registry as u

import numpy as np

# Ratio of the width of the gap between the baffle and the wall and the spacing
# between the baffles.
BAFFLE_RATIO = 1
DRAIN_TIME = 15 * u.min
MOD_ND = 0.5 * u.inch
SPACER_ND = 0.75 * u.inch
MOD_EDGE_LAST_PIPE_S = 10 * u.cm

# Height that the drain stub extends above the top of the flocculator wall
DRAIN_STUB_EXT_H = 0 * u.cm
MOD_PIPE_EDGE_S = 10 * u.cm
BAFFLE_THICKNESS = 2 * u.mm


class Flocculator:
    """Calculates physical dimensions of an AguaClara flocculator.
    Constant instance attributes
    ----------------------------
    - BAFFLE_K (K or K_{baffle}): float
        - The minor loss coefficient of the flocculator baffles.
    - CHANNEL_N_MIN (N_{channel}: int
        - The minimum number of flocculator channels.
    - HS_RATIO_MIN (\Pi_{HS}): float
        - The minimum ratio between expansion height and baffle spacing
    - RATIO_MAX_HS (\Pi_{HS}): float
        - The maximum ratio between expansion height and baffle spacing
    - SDR (sdr): float
        - The standard dimension ratio.
    """

    # Increased both to provide a safety margin on flocculator head loss and
    # to simultaneously scale back on the actual collision potential we are
    # trying to achieve.
    # Originally calculated to be 2.3 from the equations:

    # K_MINOR_FLOC_BAFFLE = (1/VC_BAFFLE_RATIO - 1)**2
    BAFFLE_K = 2.5
    CHANNEL_N_MIN = 2
    HS_RATIO_MIN = 3.0
    RATIO_MAX_HS = 6.0
    SDR = 41.0

    def __init__(
            self,
            Q=20 * u.L/u.s,
            temp=25 * u.degC,
            max_L=6 * u.m,
            Gt=37000,
            HL = 40 * u.cm,
            downstream_H = 2 * u.m,
            ent_tank_L=1.5 * u.m,
            max_W=42 * u.inch,
            drain_t=30 * u.min
    ):
        """Instantiate a Flocculator object, representing a real flocculator
        component.
        :param Q: Flow rate of water through the flocculator.
        :type Q: float * u.L/u.s
        :param temp: Water temperature of the flocculator.
        :type temp: float * u.degC
        :param max_L: Desired maximum length of the flocculator tank. Often equal to the sed tank length.
        :type max_L: float * u.m
        :param Gt: Desired total fluid deformation (collision potential).
        :type Gt: float * u.dimensionless
        :param HL: Desired head loss in the flocculator.
        :type HL: float * u.m
        :param  downstream_H: Desired depth at the downstream end.
        :type  downstream_H: float * u.m
        :param  ent_tank_L: Estimated length of entrance tank.
        :type  ent_tank_L: float * u.m
        :param  max_W: Maximum width of flocculator based on baffle sheet width.
        :type  max_W: float * u.m
        :returns: object
        :rtype: Flocculator
        """
        self.Q = Q
        self.temp = temp
        self.max_L = max_L
        self.Gt = Gt
        self.HL = HL
        self.downstream_H = downstream_H
        self.ent_tank_L = ent_tank_L
        self.max_W = max_W
        self.drain_t = drain_t

    @property
    def vel_grad_avg(self):
        """Calculate the average velocity gradient (G-bar) of water flowing
        through the flocculator.
        :returns: Average velocity gradient (G-bar)
        :rtype: float * 1 / second
        """
        return ((u.standard_gravity * self.HL) /
               (pc.viscosity_kinematic(self.temp) * self.Gt)).to(u.s ** -1)

    @property
    def retention_time(self):
        """Calculates the hydraulic retention time neglecting the volume created by head loss in the flocculator.
        :returns: Hydraulic retention time (:math:`\theta`)
        :rtype: float * second
        """
        return (self.Gt / self.vel_grad_avg).to(u.s)

    @property
    def vol(self):
        """Calculate the target volume (not counting the volume added by head loss) of the flocculator.
        :returns: Target volume
        :rtype: float * meter ** 3
        """
        return (self.Q * self.retention_time).to(u.m ** 3)

    @property
    def W_min_HS_ratio(self):
        """Calculate the minimum flocculator channel width, given the minimum
        ratio between expansion height (H) and baffle spacing (S).
        :returns: Minimum channel width given H_e/S
        :rtype: float * centimeter
        """
        return ((self.HS_RATIO_MIN * self.Q / self.downstream_H) *
               (self.BAFFLE_K /
                (2 * self.downstream_H * pc.viscosity_kinematic(self.temp) * self.vel_grad_avg ** 2)) ** (1/3)
               ).to(u.cm)

    @property
    def channel_n(self):
        """Calculate the minimum number of channels based on the maximum
        possible channel width and the maximum length of the channels.
        Round up to the next even number (factor of 2 shows up twice in equation)
        The channel width must be greater than the hydraulic width that ensure baffle overlap.
        Based on the equation for the flocculator volume
        volume = ([max_L*channel_n] - entrancetank_L)*max_W * downstream_H
        :returns: number of channels
        :rtype: float * dimensionless
        """
        min_hydraulic_W =\
            np.amax(np.array([1, (self.max_W/self.W_min_HS_ratio).to(u.dimensionless)])) * self.W_min_HS_ratio
        return 2*np.ceil(((self.vol / (min_hydraulic_W * self.downstream_H) +
                           self.ent_tank_L) / (2 * self.max_L)).to(u.dimensionless))

    @property
    def channel_W(self):
        """
        The minimum and hence optimal channel width of the flocculator.
        This
        The channel must be
        - wide enough to meet the volume requirement (channel_est_W)
        - wider than human access for construction
        - wider than hydraulic requirement to meet H/S ratio
        Create a dimensionless array of the 3 requirements and then get the maximum

        :returns: Channel width
        :rtype: float * meter
        """
        channel_est_W = (self.vol / (self.downstream_H * (self.channel_n * self.max_L - self.ent_tank_L))).to(u.m)
        # The channel may need to wider than the width that would get the exact required volume.
        # In that case we will need to shorten the flocculator
        channel_W = np.amax(np.array([1, (ha.HUMAN_W_MIN/channel_est_W).to(u.dimensionless),
                                      (self.W_min_HS_ratio/channel_est_W).to(u.dimensionless)])) * channel_est_W
        return channel_W

    @property
    def channel_L(self):
        """
        The channel length of the flocculator. If ha.HUMAN_W_MIN or W_min_HS_ratio
        is the defining constraint for the flocculator width, then the flocculator
        volume will be greater than necessary. Bring the volume back to the design
        volume by shortening the flocculator in this case. This design approach
        will produce flocculators that are the same length as the max_L that was
        specified in many cases. The flocculator will be less than the specified
        length especially for cases with only one or two sed tanks.
        channel_L = (vol/(channel_W * downstream_H) + entrancetank_L)/channel_n
        :returns: Channel length
        :rtype: float * meter
        """
        channel_L = ((self.vol / (self.channel_W * self.downstream_H) + self.ent_tank_L) / self.channel_n).to(u.m)
        return channel_L

    @property
    def expansion_max_H(self):
        """"Return the maximum distance between expansions for the largest
        allowable H/S ratio.
        :returns: Maximum expansion distance
        :rtype: float * meter
        Examples
        --------
        exp_dist_max(20*u.L/u.s, 40*u.cm, 37000, 25*u.degC, 2*u.m)
        0.375 meter
        """
        return (((self.BAFFLE_K / (2 * pc.viscosity_kinematic(self.temp) * (self.vel_grad_avg ** 2))) *
                (self.Q * self.RATIO_MAX_HS / self.channel_W) ** 3) ** (1/4)).to(u.m)

    @property
    def expansion_n(self):
        """Return the minimum number of expansions per baffle space.
        :returns: Minimum number of expansions/baffle space
        :rtype: int
        """
        return np.ceil(self.downstream_H / self.expansion_max_H)

    @property
    def expansion_H(self):
        """Returns the height between flow expansions.
        :returns: Height between flow expansions
        :rtype: float * centimeter
        """
        return (self.downstream_H / self.expansion_n).to(u.cm)

    @property
    def baffle_S(self):
        """Return the spacing between baffles.
        :returns: Spacing between baffles
        :rtype: int
        """
        return ((self.BAFFLE_K /
                ((2 * self.expansion_H * (self.vel_grad_avg ** 2) *
                 pc.viscosity_kinematic(self.temp))).to_base_units()) ** (1/3) *
               self.Q / self.channel_W).to(u.cm)

    @property
    def obstacle_n(self):
        """Return the number of obstacles per baffle.
        :returns: Number of obstacles per baffle
        :rtype: int
        """
        return self.expansion_n - 1

    @property
    def drain_K(self):
        """ Return the minor loss coefficient of the drain pipe.
        :returns: Minor Loss Coefficient
        :return: float
        """
        drain_K = minorloss.PIPE_ENTRANCE_K_MINOR + minorloss.PIPE_ENTRANCE_K_MINOR + minorloss.PIPE_EXIT_K_MINOR
        return drain_K

    @property
    def drain_D(self):
        """ Returns depth of drain pipe.
        :returns: Depth
        :return: float
        """
        tank_A = 2 * self.channel_L * self.channel_W
        drain_D = (np.sqrt(8 * tank_A / (np.pi * self.drain_t) * np.sqrt(
           self.downstream_H * self.drain_K / (2 * u.standard_gravity)))).to_base_units()
        return drain_D

    @property
    def drain_ND(self):
        """Returns the diameter of the drain pipe.
        Each drain pipe will drain two channels because channels are connected by
        a port at the far end and the first channel can't have a drain because
        of the entrance tank. Need to review the design to see if this is a good
        assumption.
        D_{Pipe} = \sqrt{ \frac{8 A_{Tank}}{\pi t_{Drain}} \sqrt{ \frac{h_0 \sum K}{2g} } }
        :returns: list of designed values
        :rtype: float * centimeter
        """
        drain_ND = pipes.ND_SDR_available(self.drain_D, self.SDR)
        return drain_ND

    @property
    def design(self):
        """Returns the designed values.
        :returns: list of designed values (G, t, channel_W, obstacle_n)
        :rtype: int
        """
        floc_dict = {'channel_n': self.channel_n,
                     'channel_L': self.channel_L,
                     'channel_W': self.channel_W,
                     'baffle_S': self.baffle_S,
                     'obstacle_n': self.obstacle_n,
                     'G': self.vel_grad_avg,
                     't': self.retention_time,
                     'expansion_max_H': self.expansion_max_H,
                     'drain_ND': self.drain_ND}
        return floc_dict

    def draw(self):
        """Draw the Onshape flocculator model based off of this object."""
        from onshapepy import Part
        CAD = Part(
            'https://cad.onshape.com/documents/b4cfd328713460beeb3125ac/w/3928b5c91bb0a0be7858d99e/e/6f2eeada21e494cebb49515f'
        )
        CAD.params = {
            'channel_L': self.channel_L,
            'channel_W': self.channel_W,
            'channel_H': self.downstream_H,
            'channel_pairs': self.channel_n/2,
            'baffle_S': self.baffle_S,
        }



"""Contains unit process functions pertaining to the design of physical
and chemical unit processes for AguaClara water treatment plants.
"""

from aguaclara.core.units import unit_registry as u
import aguaclara.core.materials as mat
import aguaclara.core.constants as con
import aguaclara.core.utility as ut

import numpy as np
from scipy import interpolate, integrate

gravity = con.GRAVITY

######################Air################################
def density_air(Pressure, MolarMass, Temperature):
    """Return the density of the air.

    :param Pressure: pressure of the air in the system
    :type Pressure: float
    :param MolarMass: molar mass of the air in the system
    :type MolarMass: float
    :param Temperature: Temperature of the air in the system
    :type Temperature: float

    :return: density of the air in the system
    :rtype: float
    """
    return (Pressure * MolarMass / (u.R * Temperature)).to(u.kg/u.m**3)

###################### Simple geometry ######################
"""A few equations for useful geometry.
Is there a geometry package that we should be using?
"""

@u.wraps(u.m**2, u.m, False)
def area_circle(DiamCircle):
    """Return the area of a circle."""
    ut.check_range([DiamCircle, ">0", "DiamCircle"])
    return np.pi / 4 * DiamCircle**2


@u.wraps(u.m, u.m**2, False)
def diam_circle(AreaCircle):
    """Return the diameter of a circle."""
    ut.check_range([AreaCircle, ">0", "AreaCircle"])
    return np.sqrt(4 * AreaCircle / np.pi)

######################### Hydraulics #########################

RE_TRANSITION_PIPE = 2100
""" """
K_KOZENY = con.K_KOZENY

WATER_DENSITY_TABLE = [(273.15, 278.15, 283.15, 293.15, 303.15, 313.15,
                        323.15, 333.15, 343.15, 353.15, 363.15, 373.15
                        ), (999.9, 1000, 999.7, 998.2, 995.7, 992.2,
                            988.1, 983.2, 977.8, 971.8, 965.3, 958.4
                            )
                       ]
"""Table of temperatures and the corresponding water density.

Index[0] is a list of water temperatures, in Kelvin.
Index[1] is the corresponding densities, in kg/m³.
"""


@u.wraps(u.kg/(u.m*u.s), [u.degK], False)
def viscosity_dynamic(temp):
    """Return the dynamic viscosity of water at a given temperature.

    If given units, the function will automatically convert to Kelvin.
    If not given units, the function will assume Kelvin.
    """
    ut.check_range([temp, ">0", "Temperature in Kelvin"])
    return 2.414 * (10**-5) * 10**(247.8 / (temp-140))


@u.wraps(u.kg/u.m**3, [u.degK], False)
def density_water(temp):
    """Return the density of water at a given temperature.

    If given units, the function will automatically convert to Kelvin.
    If not given units, the function will assume Kelvin.
    """
    ut.check_range([temp, ">0", "Temperature in Kelvin"])
    rhointerpolated = interpolate.CubicSpline(WATER_DENSITY_TABLE[0],
                                                    WATER_DENSITY_TABLE[1])
    return np.asscalar(rhointerpolated(temp))


@u.wraps(u.m**2/u.s, [u.degK], False)
def viscosity_kinematic(temp):
    """Return the kinematic viscosity of water at a given temperature.

    If given units, the function will automatically convert to Kelvin.
    If not given units, the function will assume Kelvin.
    """
    ut.check_range([temp, ">0", "Temperature in Kelvin"])
    return (viscosity_dynamic(temp).magnitude
            / density_water(temp).magnitude)


@u.wraps(None, [u.m**3/u.s, u.m, u.m**2/u.s], False)
def re_pipe(FlowRate, Diam, Nu):
    """Return the Reynolds Number for a pipe."""
    #Checking input validity
    ut.check_range([FlowRate, ">0", "Flow rate"], [Diam, ">0", "Diameter"],
                   [Nu, ">0", "Nu"])
    return (4 * FlowRate) / (np.pi * Diam * Nu)


@u.wraps(u.m, [u.m, u.m], False)
@ut.list_handler()
def radius_hydraulic(Width, DistCenter, openchannel):
    """Return the hydraulic radius.

    Width and DistCenter are length values and openchannel is a boolean.
    """
    ut.check_range([Width, ">0", "Width"], [DistCenter, ">0", "DistCenter"],
                   [openchannel, "boolean", "openchannel"])
    if openchannel:
        return (Width*DistCenter) / (Width + 2*DistCenter)
        # if openchannel is True, the channel is open. Otherwise, the channel
        # is assumed to have a top.
    else:
        return (Width*DistCenter) / (2 * (Width+DistCenter))


@u.wraps(u.m, [u.m**2, u.m], False)
def radius_hydraulic_general(Area, PerimWetted):
    """Return the general hydraulic radius."""
    ut.check_range([Area, ">0", "Area"], [PerimWetted, ">0", "Wetted perimeter"])
    return Area / PerimWetted


@u.wraps(None, [u.m**3/u.s, u.m, u.m, u.m**2/u.s], False)
def re_rect(FlowRate, Width, DistCenter, Nu, openchannel):
    """Return the Reynolds Number for a rectangular channel."""
    #Checking input validity - inputs not checked here are checked by
    #functions this function calls.
    ut.check_range([FlowRate, ">0", "Flow rate"], [Nu, ">0", "Nu"])
    return (4 * FlowRate
            * radius_hydraulic(Width, DistCenter, openchannel).magnitude
            / (Width * DistCenter * Nu))
    #Reynolds Number for rectangular channel; open = False if all sides
    #are wetted; l = Diam and Diam = 4*R.h


@u.wraps(None, [u.m/u.s, u.m**2, u.m, u.m**2/u.s], False)
def re_general(Vel, Area, PerimWetted, Nu):
    """Return the Reynolds Number for a general cross section."""
    #Checking input validity - inputs not checked here are checked by
    #functions this function calls.
    ut.check_range([Vel, ">=0", "Velocity"], [Nu, ">0", "Nu"])
    return 4 * radius_hydraulic_general(Area, PerimWetted).magnitude * Vel / Nu


@u.wraps(None, [u.m**3/u.s, u.m, u.m**2/u.s, u.m], False)
def fric(FlowRate, Diam, Nu, PipeRough):
    """Return the friction factor for pipe flow.

    This equation applies to both laminar and turbulent flows.
    """
    #Checking input validity - inputs not checked here are checked by
    #functions this function calls.
    ut.check_range([PipeRough, "0-1", "Pipe roughness"])
    if re_pipe(FlowRate, Diam, Nu) >= RE_TRANSITION_PIPE:
        #Swamee-Jain friction factor for turbulent flow; best for
        #Re>3000 and ε/Diam < 0.02
        f = (0.25 / (np.log10(PipeRough / (3.7 * Diam)
                              + 5.74 / re_pipe(FlowRate, Diam, Nu) ** 0.9
                              )
                     ) ** 2
             )
    else:
        f = 64 / re_pipe(FlowRate, Diam, Nu)
    return f


@u.wraps(None, [u.m**3/u.s, u.m, u.m, u.m**2/u.s, u.m], False)
@ut.list_handler()
def fric_rect(FlowRate, Width, DistCenter, Nu, PipeRough, openchannel):
    """Return the friction factor for a rectangular channel."""
    #Checking input validity - inputs not checked here are checked by
    #functions this function calls.
    ut.check_range([PipeRough, "0-1", "Pipe roughness"])
    if re_rect(FlowRate,Width,DistCenter,Nu,openchannel) >= RE_TRANSITION_PIPE:
        #Swamee-Jain friction factor adapted for rectangular channel.
        #Diam = 4*R_h in this case.
        return (0.25
                / (np.log10((PipeRough
                             / (3.7 * 4
                                * radius_hydraulic(Width, DistCenter,
                                                   openchannel).magnitude
                                )
                             )
                            + (5.74 / (re_rect(FlowRate, Width, DistCenter,
                                               Nu, openchannel) ** 0.9)
                               )
                            )
                    ) ** 2
                )
    else:
        return 64 / re_rect(FlowRate, Width, DistCenter, Nu, openchannel)


@u.wraps(None, [u.m**2, u.m, u.m/u.s, u.m**2/u.s, u.m], False)
@ut.list_handler()
def fric_general(Area, PerimWetted, Vel, Nu, PipeRough):
    """Return the friction factor for a general channel."""
    #Checking input validity - inputs not checked here are checked by
    #functions this function calls.
    ut.check_range([PipeRough, "0-1", "Pipe roughness"])
    if re_general(Vel, Area, PerimWetted, Nu) >= RE_TRANSITION_PIPE:
        #Swamee-Jain friction factor adapted for any cross-section.
        #Diam = 4*R*h
        f= (0.25 /
            (np.log10((PipeRough
                       / (3.7 * 4
                          * radius_hydraulic_general(Area, PerimWetted).magnitude
                          )
                       )
                      + (5.74
                         / re_general(Vel, Area, PerimWetted, Nu) ** 0.9
                         )
                      )
             ) ** 2
            )
    else:
        f = 64 / re_general(Vel, Area, PerimWetted, Nu)
    return f


@u.wraps(u.m, [u.m**3/u.s, u.m, u.m, u.m**2/u.s, u.m], False)
def headloss_fric(FlowRate, Diam, Length, Nu, PipeRough):
    """Return the major head loss (due to wall shear) in a pipe.

    This equation applies to both laminar and turbulent flows.
    """
    #Checking input validity - inputs not checked here are checked by
    #functions this function calls.
    ut.check_range([Length, ">0", "Length"])
    return (fric(FlowRate, Diam, Nu, PipeRough)
            * 8 / (gravity.magnitude * np.pi**2)
            * (Length * FlowRate**2) / Diam**5
            )


@u.wraps(u.m, [u.m**3/u.s, u.m], False)
def headloss_exp(FlowRate, Diam, KMinor):
    """Return the minor head loss (due to expansions) in a pipe.

    This equation applies to both laminar and turbulent flows.
    """
    #Checking input validity
    ut.check_range([FlowRate, ">0", "Flow rate"], [Diam, ">0", "Diameter"],
                   [KMinor, ">=0", "K minor"])
    return KMinor * 8 / (gravity.magnitude * np.pi**2) * FlowRate**2 / Diam**4


@u.wraps(u.m, [u.m**3/u.s, u.m, u.m, u.m**2/u.s, u.m], False)
def headloss(FlowRate, Diam, Length, Nu, PipeRough, KMinor):
    """Return the total head loss from major and minor losses in a pipe.

    This equation applies to both laminar and turbulent flows.
    """
    #Inputs do not need to be checked here because they are checked by
    #functions this function calls.
    return (headloss_fric(FlowRate, Diam, Length, Nu, PipeRough).magnitude
            + headloss_exp(FlowRate, Diam, KMinor).magnitude)


@u.wraps(u.m, [u.m**3/u.s, u.m, u.m, u.m, u.m**2/u.s, u.m], False)
def headloss_fric_rect(FlowRate, Width, DistCenter, Length, Nu, PipeRough, openchannel):
    """Return the major head loss due to wall shear in a rectangular channel.

    This equation applies to both laminar and turbulent flows.
    """
    #Checking input validity - inputs not checked here are checked by
    #functions this function calls.
    ut.check_range([Length, ">0", "Length"])
    return (fric_rect(FlowRate, Width, DistCenter, Nu,
                      PipeRough, openchannel)
            * Length
            / (4 * radius_hydraulic(Width, DistCenter, openchannel).magnitude)
            * FlowRate**2
            / (2 * gravity.magnitude * (Width*DistCenter)**2)
            )


@u.wraps(u.m, [u.m**3/u.s, u.m, u.m], False)
def headloss_exp_rect(FlowRate, Width, DistCenter, KMinor):
    """Return the minor head loss due to expansion in a rectangular channel.

    This equation applies to both laminar and turbulent flows.
    """
    #Checking input validity
    ut.check_range([FlowRate, ">0", "Flow rate"], [Width, ">0", "Width"],
                   [DistCenter, ">0", "DistCenter"], [KMinor, ">=0", "K minor"])
    return (KMinor * FlowRate**2
            / (2 * gravity.magnitude * (Width*DistCenter)**2)
            )


@u.wraps(u.m, [u.m**3/u.s, u.m, u.m, u.m, None, u.m**2/u.s, u.m], False)
def headloss_rect(FlowRate, Width, DistCenter, Length,
                  KMinor, Nu, PipeRough, openchannel):
    """Return the total head loss in a rectangular channel.

    Total head loss is a combination of the major and minor losses.
    This equation applies to both laminar and turbulent flows.
    """
    #Inputs do not need to be checked here because they are checked by
    #functions this function calls.
    return (headloss_exp_rect(FlowRate, Width, DistCenter, KMinor).magnitude
              + headloss_fric_rect(FlowRate, Width, DistCenter, Length,
                                   Nu, PipeRough, openchannel).magnitude)


@u.wraps(u.m, [u.m**2, u.m, u.m/u.s, u.m, u.m**2/u.s, u.m], False)
def headloss_fric_general(Area, PerimWetted, Vel, Length, Nu, PipeRough):
    """Return the major head loss due to wall shear in the general case.

    This equation applies to both laminar and turbulent flows.
    """
    #Checking input validity - inputs not checked here are checked by
    #functions this function calls.
    ut.check_range([Length, ">0", "Length"])
    return (fric_general(Area, PerimWetted, Vel, Nu, PipeRough) * Length
            / (4 * radius_hydraulic_general(Area, PerimWetted).magnitude)
            * Vel**2 / (2*gravity.magnitude)
            )


@u.wraps(u.m, [u.m/u.s], False)
def headloss_exp_general(Vel, KMinor):
    """Return the minor head loss due to expansion in the general case.

    This equation applies to both laminar and turbulent flows.
    """
    #Checking input validity
    ut.check_range([Vel, ">0", "Velocity"], [KMinor, '>=0', 'K minor'])
    return KMinor * Vel**2 / (2*gravity.magnitude)


@u.wraps(u.m, [u.m**2, u.m/u.s, u.m, u.m, None, u.m**2/u.s, u.m], False)
def headloss_gen(Area, Vel, PerimWetted, Length, KMinor, Nu, PipeRough):
    """Return the total head lossin the general case.

    Total head loss is a combination of major and minor losses.
    This equation applies to both laminar and turbulent flows.
    """
    #Inputs do not need to be checked here because they are checked by
    #functions this function calls.
    return (headloss_exp_general(Vel, KMinor).magnitude
            + headloss_fric_general(Area, PerimWetted, Vel,
                                     Length, Nu, PipeRough).magnitude)


@u.wraps(u.m, [u.m**3/u.s, u.m, u.m, None,
               u.m**2/u.s, u.m], False)
def headloss_manifold(FlowRate, Diam, Length, KMinor, Nu, PipeRough, NumOutlets):
    """Return the total head loss through the manifold."""
    #Checking input validity - inputs not checked here are checked by
    #functions this function calls.
    ut.check_range([NumOutlets, ">0, int", 'Number of outlets'])
    return (headloss(FlowRate, Diam, Length, Nu, PipeRough, KMinor).magnitude
            * ((1/3 )
               + (1 / (2*NumOutlets))
               + (1 / (6*NumOutlets**2))
               )
            )


@u.wraps(u.m**3/u.s, [u.m, u.m], False)
@ut.list_handler()
def flow_orifice(Diam, Height, RatioVCOrifice):
    """Return the flow rate of the orifice."""
    #Checking input validity
    ut.check_range([Diam, ">0", "Diameter"],
                   [RatioVCOrifice, "0-1", "VC orifice ratio"])
    if Height > 0:
        return (RatioVCOrifice * area_circle(Diam).magnitude
                * np.sqrt(2 * gravity.magnitude * Height))
    else:
        return 0


#Deviates from the MathCad at the 6th decimal place. Worth investigating or not?
@u.wraps(u.m**3/u.s, [u.m, u.m], False)
@ut.list_handler()
def flow_orifice_vert(Diam, Height, RatioVCOrifice):
    """Return the vertical flow rate of the orifice."""
    #Checking input validity
    ut.check_range([RatioVCOrifice, "0-1", "VC orifice ratio"])
    if Height > -Diam / 2:
        flow_vert = integrate.quad(lambda z: (Diam * np.sin(np.arccos(z/(Diam/2)))
                                                   * np.sqrt(Height - z)
                                                   ),
                                                   - Diam / 2,
                                                   min(Diam/2, Height))
        return flow_vert[0] * RatioVCOrifice * np.sqrt(2 * gravity.magnitude)
    else:
        return 0


@u.wraps(u.m, [u.m, None, u.m**3/u.s], False)
def head_orifice(Diam, RatioVCOrifice, FlowRate):
    """Return the head of the orifice."""
    #Checking input validity
    ut.check_range([Diam, ">0", "Diameter"], [FlowRate, ">0", "Flow rate"],
                   [RatioVCOrifice, "0-1", "VC orifice ratio"])
    return ((FlowRate
             / (RatioVCOrifice * area_circle(Diam).magnitude)
             )**2
            / (2*gravity.magnitude)
            )


@u.wraps(u.m**2, [u.m, None, u.m**3/u.s], False)
def area_orifice(Height, RatioVCOrifice, FlowRate):
    """Return the area of the orifice."""
    #Checking input validity
    ut.check_range([Height, ">0", "Height"], [FlowRate, ">0", "Flow rate"],
                   [RatioVCOrifice, "0-1, >0", "VC orifice ratio"])
    return FlowRate / (RatioVCOrifice * np.sqrt(2 * gravity.magnitude * Height))


@u.wraps(None, [u.m**3/u.s, None, u.m, u.m], False)
def num_orifices(FlowPlant, RatioVCOrifice, HeadLossOrifice, DiamOrifice):
    """Return the number of orifices."""
    #Inputs do not need to be checked here because they are checked by
    #functions this function calls.
    return np.ceil(area_orifice(HeadLossOrifice, RatioVCOrifice,
                                 FlowPlant).magnitude
                    / area_circle(DiamOrifice).magnitude)


# Here we define functions that return the flow rate.
@u.wraps(u.m**3/u.s, [u.m, u.m**2/u.s], False)
def flow_transition(Diam, Nu):
    """Return the flow rate for the laminar/turbulent transition.

    This equation is used in some of the other equations for flow.
    """
    #Checking input validity
    ut.check_range([Diam, ">0", "Diameter"], [Nu, ">0", "Nu"])
    return np.pi * Diam * RE_TRANSITION_PIPE * Nu / 4


@u.wraps(u.m**3/u.s, [u.m, u.m, u.m, u.m**2/u.s], False)
def flow_hagen(Diam, HeadLossFric, Length, Nu):
    """Return the flow rate for laminar flow with only major losses."""
    #Checking input validity
    ut.check_range([Diam, ">0", "Diameter"], [Length, ">0", "Length"],
                   [HeadLossFric, ">=0", "Headloss due to friction"],
                   [Nu, ">0", "Nu"])
    return (np.pi*Diam**4) / (128*Nu) * gravity.magnitude * HeadLossFric / Length


@u.wraps(u.m**3/u.s, [u.m, u.m, u.m, u.m**2/u.s, u.m], False)
def flow_swamee(Diam, HeadLossFric, Length, Nu, PipeRough):
    """Return the flow rate for turbulent flow with only major losses."""
    #Checking input validity
    ut.check_range([Diam, ">0", "Diameter"], [Length, ">0", "Length"],
                   [HeadLossFric, ">0", "Headloss due to friction"],
                   [Nu, ">0", "Nu"], [PipeRough, "0-1", "Pipe roughness"])
    logterm = np.log10(PipeRough / (3.7 * Diam)
                       + 2.51 * Nu * np.sqrt(Length / (2 * gravity.magnitude
                                                         * HeadLossFric
                                                         * Diam**3)
                                              )
                       )
    return ((-np.pi / np.sqrt(2)) * Diam**(5/2) * logterm
            * np.sqrt(gravity.magnitude * HeadLossFric / Length)
            )


@u.wraps(u.m**3/u.s, [u.m, u.m, u.m, u.m**2/u.s, u.m], False)
@ut.list_handler()
def flow_pipemajor(Diam, HeadLossFric, Length, Nu, PipeRough):
    """Return the flow rate with only major losses.

    This function applies to both laminar and turbulent flows.
    """
    #Inputs do not need to be checked here because they are checked by
    #functions this function calls.
    FlowHagen = flow_hagen(Diam, HeadLossFric, Length, Nu).magnitude
    if FlowHagen < flow_transition(Diam, Nu).magnitude:
        return FlowHagen
    else:
        return flow_swamee(Diam, HeadLossFric, Length, Nu, PipeRough).magnitude


@u.wraps(u.m**3/u.s, [u.m, u.m], False)
def flow_pipeminor(Diam, HeadLossExpans, KMinor):
    """Return the flow rate with only minor losses.

    This function applies to both laminar and turbulent flows.
    """
    #Checking input validity - inputs not checked here are checked by
    #functions this function calls.
    ut.check_range([HeadLossExpans, ">=0", "Headloss due to expansion"],
                   [KMinor, ">0", "K minor"])
    return (area_circle(Diam).magnitude * np.sqrt(2 * gravity.magnitude
                                                  * HeadLossExpans
                                                  / KMinor)
            )

# Now we put all of the flow equations together and calculate the flow in a
# straight pipe that has both major and minor losses and might be either
# laminar or turbulent.
@u.wraps(u.m**3/u.s, (u.m, u.m, u.m, u.m**2/u.s, u.m), False)
@ut.list_handler()
def flow_pipe(Diam, HeadLoss, Length, Nu, PipeRough, KMinor):
    """Return the the flow in a straight pipe.

    This function works for both major and minor losses and
    works whether the flow is laminar or turbulent.
    """
    #Inputs do not need to be checked here because they are checked by
    #functions this function calls.
    if KMinor == 0:
        FlowRate = flow_pipemajor(Diam, HeadLoss, Length, Nu,
                                  PipeRough).magnitude
    else:
        FlowRatePrev = 0
        err = 1.0
        FlowRate = min(flow_pipemajor(Diam, HeadLoss, Length,
                                      Nu, PipeRough).magnitude,
                       flow_pipeminor(Diam, HeadLoss, KMinor).magnitude
                       )
        while err > 0.01:
            FlowRatePrev = FlowRate
            HLFricNew = (HeadLoss * headloss_fric(FlowRate, Diam, Length,
                                                  Nu, PipeRough).magnitude
                         / (headloss_fric(FlowRate, Diam, Length,
                                          Nu, PipeRough).magnitude
                            + headloss_exp(FlowRate, Diam, KMinor).magnitude
                            )
                         )
            FlowRate = flow_pipemajor(Diam, HLFricNew, Length,
                                      Nu, PipeRough).magnitude
            if FlowRate == 0:
                err = 0.0
            else:
                err = (abs(FlowRate - FlowRatePrev)
                       / ((FlowRate + FlowRatePrev) / 2)
                       )
    return FlowRate


@u.wraps(u.m, [u.m**3/u.s, u.m, u.m, u.m**2/u.s], False)
def diam_hagen(FlowRate, HeadLossFric, Length, Nu):
    #Checking input validity
    ut.check_range([FlowRate, ">0", "Flow rate"], [Length, ">0", "Length"],
                   [HeadLossFric, ">0", "Headloss due to friction"],
                   [Nu, ">0", "Nu"])
    return ((128 * Nu * FlowRate * Length)
            / (gravity.magnitude * HeadLossFric * np.pi)
            ) ** (1/4)


@u.wraps(u.m, [u.m**3/u.s, u.m, u.m, u.m**2/u.s, u.m], False)
def diam_swamee(FlowRate, HeadLossFric, Length, Nu, PipeRough):
    """Return the inner diameter of a pipe.

    The Swamee Jain equation is dimensionally correct and returns the
    inner diameter of a pipe given the flow rate and the head loss due
    to shear on the pipe walls. The Swamee Jain equation does NOT take
    minor losses into account. This equation ONLY applies to turbulent
    flow.
    """
    #Checking input validity
    ut.check_range([FlowRate, ">0", "Flow rate"], [Length, ">0", "Length"],
                   [HeadLossFric, ">0", "Headloss due to friction"],
                   [Nu, ">0", "Nu"], [PipeRough, "0-1", "Pipe roughness"])
    a = ((PipeRough ** 1.25)
         * ((Length * FlowRate**2)
            / (gravity.magnitude * HeadLossFric)
            )**4.75
         )
    b = (Nu * FlowRate**9.4
         * (Length / (gravity.magnitude *  HeadLossFric)) ** 5.2
         )
    return 0.66 * (a+b)**0.04


@u.wraps(u.m, [u.m**3/u.s, u.m, u.m, u.m**2/u.s, u.m], False)
@ut.list_handler()
def diam_pipemajor(FlowRate, HeadLossFric, Length, Nu, PipeRough):
    """Return the pipe IDiam that would result in given major losses.
    This function applies to both laminar and turbulent flow.
    """
    #Inputs do not need to be checked here because they are checked by
    #functions this function calls.
    DiamLaminar = diam_hagen(FlowRate, HeadLossFric, Length, Nu).magnitude
    if re_pipe(FlowRate, DiamLaminar, Nu) <= RE_TRANSITION_PIPE:
        return DiamLaminar
    else:
        return diam_swamee(FlowRate, HeadLossFric, Length,
                           Nu, PipeRough).magnitude

@u.wraps(u.m, [u.m**3/u.s, u.m], False)
def diam_pipeminor(FlowRate, HeadLossExpans, KMinor):
    """Return the pipe ID that would result in the given minor losses.

    This function applies to both laminar and turbulent flow.
    """
    #Checking input validity
    ut.check_range([FlowRate, ">0", "Flow rate"], [KMinor, ">=0", "K minor"],
                   [HeadLossExpans, ">0", "Headloss due to expansion"])
    return (np.sqrt(4 * FlowRate / np.pi)
            * (KMinor / (2 * gravity.magnitude * HeadLossExpans)) ** (1/4)
            )


@u.wraps(u.m, [u.m**3/u.s, u.m, u.m, u.m**2/u.s, u.m], False)
@ut.list_handler()
def diam_pipe(FlowRate, HeadLoss, Length, Nu, PipeRough, KMinor):
    """Return the pipe ID that would result in the given total head loss.

    This function applies to both laminar and turbulent flow and
    incorporates both minor and major losses.
    """
    #Inputs do not need to be checked here because they are checked by
    #functions this function calls.
    if KMinor == 0:
        Diam = diam_pipemajor(FlowRate, HeadLoss, Length, Nu,
                              PipeRough).magnitude
    else:
        Diam = max(diam_pipemajor(FlowRate, HeadLoss,
                                  Length, Nu, PipeRough).magnitude,
                   diam_pipeminor(FlowRate, HeadLoss, KMinor).magnitude)
        err = 1.00
        while err > 0.001:
            DiamPrev = Diam
            HLFricNew = (HeadLoss * headloss_fric(FlowRate, Diam, Length,
                                                  Nu, PipeRough
                                                  ).magnitude
                         / (headloss_fric(FlowRate, Diam, Length,
                                          Nu, PipeRough
                                          ).magnitude
                                          + headloss_exp(FlowRate,
                                                         Diam, KMinor
                                                         ).magnitude
                            )
                         )
            Diam = diam_pipemajor(FlowRate, HLFricNew, Length, Nu, PipeRough
                                  ).magnitude
            err = abs(Diam - DiamPrev) / ((Diam + DiamPrev) / 2)
    return Diam

# Weir head loss equations
@u.wraps(u.m, [u.m**3/u.s, u.m], False)
def width_rect_weir(FlowRate, Height):
    """Return the width of a rectangular weir."""
    #Checking input validity
    ut.check_range([FlowRate, ">0", "Flow rate"], [Height, ">0", "Height"])
    return ((3 / 2) * FlowRate
            / (con.VC_ORIFICE_RATIO * np.sqrt(2 * gravity.magnitude) * Height ** (3 / 2))
            )


# For a pipe, Width is the circumference of the pipe.
# Head loss for a weir is the difference in height between the water
# upstream of the weir and the top of the weir.
@u.wraps(u.m, [u.m**3/u.s, u.m], False)
def headloss_weir(FlowRate, Width):
    """Return the headloss of a weir."""
    #Checking input validity
    ut.check_range([FlowRate, ">0", "Flow rate"], [Width, ">0", "Width"])
    return (((3/2) * FlowRate
             / (con.VC_ORIFICE_RATIO * np.sqrt(2 * gravity.magnitude) * Width)
             ) ** (2/3))


@u.wraps(u.m, [u.m, u.m], False)
def flow_rect_weir(Height, Width):
    """Return the flow of a rectangular weir."""
    #Checking input validity
    ut.check_range([Height, ">0", "Height"], [Width, ">0", "Width"])
    return ((2/3) * con.VC_ORIFICE_RATIO
            * (np.sqrt(2*gravity.magnitude) * Height**(3/2))
            * Width)


@u.wraps(u.m, [u.m**3/u.s, u.m], False)
def height_water_critical(FlowRate, Width):
    """Return the critical local water depth."""
    #Checking input validity
    ut.check_range([FlowRate, ">0", "Flow rate"], [Width, ">0", "Width"])
    return (FlowRate / (Width * np.sqrt(gravity.magnitude))) ** (2/3)


@u.wraps(u.m/u.s, u.m, False)
def vel_horizontal(HeightWaterCritical):
    """Return the horizontal velocity."""
    #Checking input validity
    ut.check_range([HeightWaterCritical, ">0", "Critical height of water"])
    return np.sqrt(gravity.magnitude * HeightWaterCritical)


@u.wraps(u.m, [u.m, u.m, u.m/u.s, u.m, u.m**2/u.s], False)
def headloss_kozeny(Length, Diam, Vel, Porosity, Nu):
    """Return the Carmen Kozeny Sand Bed head loss."""
    #Checking input validity
    ut.check_range([Length, ">0", "Length"], [Diam, ">0", "Diam"],
                   [Vel, ">0", "Velocity"], [Nu, ">0", "Nu"],
                   [Porosity, "0-1", "Porosity"])
    return (K_KOZENY * Length * Nu
            / gravity.magnitude * (1-Porosity)**2
            / Porosity**3 * 36 * Vel
            / Diam ** 2)



from aguaclara.core.cache import ac_cache, HashableObject


class ComputedObject(HashableObject):
    def __init__(self):
        self.a = 2
        self.b = 3
        self.c = 4

    @property
    @ac_cache
    def product(self):
        increment_n_calls()
        return self.a * self.b * self.c

    @property
    @ac_cache
    def sum(self):
        increment_n_calls()
        return self.a + self.b + self.c

    @ac_cache
    def sum_with_arg(self, my_arg):
        increment_n_calls()
        return self.sum + my_arg

    @ac_cache
    def sum_with_kwarg(self, my_arg=10):
        increment_n_calls()
        return self.sum + my_arg


# Keep track of the total number of calls
side_effect_n_calls = 0


def increment_n_calls():
    global side_effect_n_calls
    side_effect_n_calls = side_effect_n_calls +1


def test_ac_cache():
    my_computed_object = ComputedObject()
    assert 24 == my_computed_object.product
    assert 1 == side_effect_n_calls
    assert 9 == my_computed_object.sum
    assert 2 == side_effect_n_calls
    assert 9 == my_computed_object.sum
    assert 2 == side_effect_n_calls
    my_computed_object.a=3
    assert 36 == my_computed_object.product
    assert 3 == side_effect_n_calls
    assert 10 == my_computed_object.sum
    assert 4 == side_effect_n_calls
    assert 15 == my_computed_object.sum_with_arg(5)
    assert 5 == side_effect_n_calls
    assert 20 == my_computed_object.sum_with_kwarg()
    assert 6 == side_effect_n_calls
    assert 20 == my_computed_object.sum_with_kwarg()
    assert 6 == side_effect_n_calls
    assert 25 == my_computed_object.sum_with_kwarg(my_arg=15)
    assert 7 == side_effect_n_calls



import unittest

from aguaclara.play import *


class UtilityTest(unittest.TestCase):

    def test_round_sf(self):
        self.assertAlmostEqual(ut.round_sf(123456.789, 8), 123456.79)
        self.assertAlmostEqual(ut.round_sf(20.01 * u.L/u.s, 2), 20 * u.L/u.s)
        self.assertAlmostEqual(ut.round_sf(-456.789 * u.L/u.s, 4), -456.8 * u.L/u.s)
        self.assertAlmostEqual(ut.round_sf(0, 4), 0)
        self.assertAlmostEqual(ut.round_sf(0 * u.m, 4), 0 * u.m)



"""This file is used to calculate the design paramters for the chemical dose
controller of an AguaClara plant.

"""

import numpy as np

from aguaclara.core import physchem as pc, utility as ut

from aguaclara.core.units import unit_registry as u


#==============================================================================
# Functions for Coagulant Viscosities and Selecting Available Tube Diameters
#==============================================================================


def _DiamTubeAvail(en_tube_series = True):
    if en_tube_series:
        return 1*u.mm
    else:
        return (1/16)*u.inch
# This section may be unnecessary
#==============================================================================
# def _DiamTubeAvail(en_tube_series = True):
#     if en_tube_series:
#         return 1*u.mm
#     else:
#         return (1/16)*u.inch
#==============================================================================

@u.wraps(u.m**2/u.s, [u.kg/u.m**3, u.degK], False)
def viscosity_kinematic_alum(conc_alum, temp):
    """Return the dynamic viscosity of water at a given temperature.

    If given units, the function will automatically convert to Kelvin.
    If not given units, the function will assume Kelvin.
    This function assumes that the temperature dependence can be explained
    based on the effect on water and that there is no confounding effect from
    the coagulant.
    """
    nu = (1 + (4.255 * 10**-6) * conc_alum**2.289) * pc.viscosity_kinematic(temp).magnitude
    return nu


@u.wraps(u.m**2/u.s, [u.kg/u.m**3, u.degK], False)
def viscosity_kinematic_pacl(conc_pacl, temp):
    """Return the dynamic viscosity of water at a given temperature.

    If given units, the function will automatically convert to Kelvin.
    If not given units, the function will assume Kelvin.
    This function assumes that the temperature dependence can be explained
    based on the effect on water and that there is no confounding effect from
    the coagulant.
    """
    nu = (1 + (2.383 * 10**-5) * conc_pacl**1.893) * pc.viscosity_kinematic(temp).magnitude
    return nu


@u.wraps(u.m**2/u.s, [u.kg/u.m**3, u.degK, None], False)
def viscosity_kinematic_chem(conc_chem, temp, en_chem):
     """Return the dynamic viscosity of water at a given temperature.

    If given units, the function will automatically convert to Kelvin.
    If not given units, the function will assume Kelvin.
    """
     if en_chem == 0:
         nu = viscosity_kinematic_alum(conc_chem, temp).magnitude
     if en_chem == 1:
         nu =  viscosity_kinematic_pacl(conc_chem, temp).magnitude
     if en_chem not in [0,1]:
         nu =  pc.viscosity_kinematic(temp).magnitude
     return nu



#==============================================================================
# Flow rate Constraints for Laminar Tube Flow, Deviation from Linear Head Loss
# Behavior, and Lowest Possible Flow
#==============================================================================


@u.wraps(u.m**3/u.s, [u.m, u.m, None, None], False)
def max_linear_flow(Diam, HeadlossCDC, Ratio_Error, KMinor):
    """Return the maximum flow that will meet the linear requirement.
    Maximum flow that can be put through a tube of a given diameter without
    exceeding the allowable deviation from linear head loss behavior
    """
    flow = (pc.area_circle(Diam)).magnitude * np.sqrt((2 * Ratio_Error * HeadlossCDC * pc.gravity)/ KMinor)
    return flow.magnitude




#==============================================================================
# Length of Tubing Required Given Head Loss, Max Flow, and Diameter
#==============================================================================

# Length of tube required to get desired head loss at maximum flow based on
# the Hagen-Poiseuille equation.
@u.wraps(u.m, [u.m**3/u.s, u.m, u.m, u.kg/u.m**3, u.degK, None, None], False)
def _len_tube(Flow, Diam, HeadLoss, conc_chem, temp, en_chem, KMinor):
    """Length of tube required to get desired head loss at maximum flow based on
    the Hagen-Poiseuille equation."""
    num1 = pc.gravity.magnitude * HeadLoss * np.pi * (Diam**4)
    denom1 = 128 * viscosity_kinematic_chem(conc_chem, temp, en_chem) * Flow
    num2 = Flow * KMinor
    denom2 = 16 * np.pi * viscosity_kinematic_chem(conc_chem, temp, en_chem)
    len = ((num1/denom1) - (num2/denom2))
    return len.magnitude



#==============================================================================
# Helper Functions
#==============================================================================
@u.wraps(None, [u.m**3/u.s, u.kg/u.m**3, u.kg/u.m**3, u.m, u.m, None, None], False)
def _n_tube_array(FlowPlant, ConcDoseMax, ConcStock,
                  DiamTubeAvail, HeadlossCDC, Ratio_Error, KMinor):

    np.ceil((FlowPlant * ConcDoseMax
               )/ ConcStock*max_linear_flow(DiamTubeAvail, HeadlossCDC, Ratio_Error, KMinor).magnitude)
    return np.ceil((FlowPlant * ConcDoseMax) /
            (ConcStock * max_linear_flow(DiamTubeAvail, HeadlossCDC, Ratio_Error, KMinor).magnitude))


@u.wraps(u.m**3/u.s, [u.m**3/u.s, u.kg/u.m**3, u.kg/u.m**3], False)
def _flow_chem_stock(FlowPlant, ConcDoseMax, ConcStock):
    FlowPlant * ConcDoseMax / ConcStock
    return FlowPlant * ConcDoseMax / ConcStock


@u.wraps(u.m**3/u.s, [u.m**3/u.s, u.kg/u.m**3, u.kg/u.m**3, u.m, u.m,None,None], False)
def _flow_cdc_tube(FlowPlant, ConcDoseMax, ConcStock,
                   DiamTubeAvail, HeadlossCDC,Ratio_Error, KMinor):

    (_flow_chem_stock(FlowPlant, ConcDoseMax, ConcStock)
     ) / (_n_tube_array(FlowPlant, ConcDoseMax, ConcStock,
                  DiamTubeAvail, HeadlossCDC,Ratio_Error, KMinor))
    return (_flow_chem_stock(FlowPlant, ConcDoseMax, ConcStock).magnitude
            ) / (_n_tube_array(FlowPlant, ConcDoseMax, ConcStock,
                        DiamTubeAvail, HeadlossCDC,Ratio_Error, KMinor))



#
@u.wraps(u.m, [u.m**3/u.s, u.kg/u.m**3, u.kg/u.m**3, u.m, u.m, u.degK, None, None], False)
def _length_cdc_tube_array(FlowPlant, ConcDoseMax, ConcStock,
                           DiamTubeAvail, HeadlossCDC, temp, en_chem, KMinor):
    """Calculate the length of each diameter tube given the corresponding flow rate
    and coagulant. Choose the tube that is shorter than the maximum length tube."""

    Flow = _flow_cdc_tube(FlowPlant, ConcDoseMax, ConcStock, DiamTubeAvail, HeadlossCDC,Ratio_Error, KMinor).magnitude


    return _len_tube(Flow, DiamTubeAvail, HeadlossCDC, ConcStock, temp, en_chem, KMinor).magnitude


# Find the index of that tube
@u.wraps(None, [u.m**3/u.s, u.kg/u.m**3, u.kg/u.m**3, u.m, u.m, u.m, u.degK, None, None], False)
def i_cdc(FlowPlant, ConcDoseMax, ConcStock,
          DiamTubeAvail, HeadlossCDC, LenCDCTubeMax, temp,
          en_chem, KMinor):

    tube_array = (_length_cdc_tube_array(FlowPlant, ConcDoseMax, ConcStock,
                                        DiamTubeAvail, HeadlossCDC, temp, en_chem,
                                        KMinor)).magnitude


    if tube_array[0] < float(LenCDCTubeMax):
        y = ut.floor_nearest(LenCDCTubeMax,tube_array)
        myindex =np.where(tube_array==y)[0][0]

    else:
        myindex = 0

    return myindex



#==============================================================================
# Final easy to use functions
#==============================================================================

@u.wraps(u.m, [u.m**3/u.s, u.kg/u.m**3, u.kg/u.m**3, u.m, u.m, u.m, u.degK, None, None], False)
def len_cdc_tube(FlowPlant, ConcDoseMax, ConcStock,
                 DiamTubeAvail, HeadlossCDC, LenCDCTubeMax, temp,
                 en_chem, KMinor):
   """The length of tubing may be longer than the max specified if the stock
   concentration is too high to give a viable solution with the specified
   length of tubing."""
   index = i_cdc(FlowPlant, ConcDoseMax, ConcStock,
                DiamTubeAvail, HeadlossCDC, LenCDCTubeMax, temp,
                en_chem, KMinor)
   len_cdc_tube = (_length_cdc_tube_array(FlowPlant, ConcDoseMax, ConcStock,
                                        DiamTubeAvail, HeadlossCDC, temp, en_chem,
                                        KMinor))[index].magnitude

   return len_cdc_tube


@u.wraps(u.m, [u.m**3/u.s, u.kg/u.m**3, u.kg/u.m**3, u.m, u.m, u.m, u.degK, None, None], False)
def diam_cdc_tube(FlowPlant, ConcDoseMax, ConcStock,
                  DiamTubeAvail, HeadlossCDC, LenCDCTubeMax,
                  temp, en_chem, KMinor):

    index = i_cdc(FlowPlant, ConcDoseMax, ConcStock,
                   DiamTubeAvail, HeadlossCDC, LenCDCTubeMax,
                   temp, en_chem, KMinor)

    diam_cdc_tube = DiamTubeAvail[index]

    return diam_cdc_tube


@u.wraps(None, [u.m**3/u.s, u.kg/u.m**3, u.kg/u.m**3, u.m, u.m, u.m, u.degK, None, None], False)
def n_cdc_tube(FlowPlant, ConcDoseMax, ConcStock,
          DiamTubeAvail, HeadlossCDC, LenCDCTubeMax,
          temp, en_chem, KMinor):

    index = i_cdc(FlowPlant, ConcDoseMax, ConcStock,
                  DiamTubeAvail, HeadlossCDC, LenCDCTubeMax,
                  temp, en_chem, KMinor)

    n_cdc_tube = _n_tube_array(FlowPlant, ConcDoseMax, ConcStock,
                  DiamTubeAvail, HeadlossCDC, Ratio_Error, KMinor)[index]

    return n_cdc_tube




# testing
FlowPlant = 100*u.L/u.s
DiamTubeAvail = np.array(np.arange(1/16,6/16,1/16))*u.inch
temp = u.Quantity(20,u.degC)
HeadlossCDC = 20*(u.cm)
ConcStock = 51.4*(u.gram/u.L)
ConcDoseMax = 2*(u.mg/u.L)
LenCDCTubeMax = 4 * u.m
en_chem = 2
KMinor = 2
Ratio_Error=0.1
x=len_cdc_tube(FlowPlant, ConcDoseMax, ConcStock,
                 DiamTubeAvail, HeadlossCDC, LenCDCTubeMax, temp,
                 en_chem, KMinor)
#print(x)
#print(diam_cdc_tube(FlowPlant, ConcDoseMax, ConcStock, DiamTubeAvail, HeadlossCDC, LenCDCTubeMax, temp, en_chem, KMinor).to(u.inch))
#print(n_cdc_tube(FlowPlant, ConcDoseMax, ConcStock, DiamTubeAvail, HeadlossCDC, LenCDCTubeMax, temp, en_chem, KMinor))
LEVER_ARM_L = 0.5 * u.m
LEVER_CYLINDER_1_D = 1 * u.inch
LEVER_CYLINDER_4_D = 2 * u.inch
LEVER_CYLINDER_2_D = 0.5 * u.inch
LEVER_CYLINDER_2_PIVOT_DIST = 6 * u.cm
CYLINDER_2_CYLINDER_3_DIST = 9.5 * u.cm
LEVER_PIVOT_BOX_L = 2 * u.inch
LEVER_PIVOT_BOX_W = 1 * u.inch
LEVER_PIVOT_BOX_H = 1 * u.inch
LEVER_ARM_THICKNESS = 0.125 * u.inch
LEVER_ARM_H = 1 * u.inch
LEVER_INNERBAR_L = 7 * u.inch
LEVER_MOUNTING_PLATE_L = 6 * u.inch
LEVER_MOUNTING_PLATE_W = 0.5 * u.cm
LEVER_MOUNTING_PLATE_H = 2 * u.inch
LEVER_TO_ENT_TANK_Z_TOP_S = 1 * u.cm
FLOAT_THICKNESS = 5 * u.cm
FLOAT_CABLE_D = 0.5 * u.cm
LEVER_SLIDER_ORIGIN_TO_SCREW_L = 1 * u.inch
LEVER_SLIDER_THICKNESS = 0.25 * u.inch
LEVER_SLIDER_H = 1.5 * u.inch
LEVER_SLIDER_L = 3 * u.inch
LEVER_SLIDER_SHORT_H = 0.125 * u.inch
LEVER_CYLINDER_L = 6 * u.inch
ENT_TANK_FRONT_WALL_TO_CDC_FLOAT_L = 0.874 * u.m

LEVER_L = 0.5 * u.m #This may be obsolete now... mrf222 2/10/16

LEVER_ARM_W = 0.0032 * u.m

LEVER_ARM_H = 0.0254 * u.m

D_CDC_CHT = 6 * u.inch

#Distance from the top of the entrance tank to the to the middle of the
# lever arm hole for the cable - (minus the) radius of the hole.
LEVER_HOLE_H = 0.0132 * u.m - (0.0095/2 * u.m)

DIAM_CABLE = 0.1 * u.inch

#Edited DLABOrigintoLAOriginZ to accommodate dimensions from McMaster
# vs Inserted Drawing
DIAM_LAB_ORIGIN_TO_LA_ORIGIN_Z = 0.0245 * u.m

#Distance from the lever arm origin to the outside center of the top part
# of the drop tube in the y direction.
LA_ORIGIN_TO_DT_Y_L = 0.7812 * u.m

#Distance from the lever arm origin to the drop tube in the z direction.
LA_ORIGIN_TO_DT_Z_L = 0.0429 * u.m

#Distance from the lever arm origin to the center of the drop tube in the
#x direction.
LA_ORIGIN_TO_DT_CENTER_X_L = 0.0290 * u.m

#Measured from CDC research team's apparatus.
CDC_REDUCER_THICKNESS = 9.5 * u.mm

#Distance from the lever arm origin to the center of the reducer in the
# x direction.
LA_ORIGIN_TO_REDUCER_X_L = 0.0290 * u.m

#Distance from the lever arm origin to the outside center of the top part
# of the reducer in the y direction.
LA_ORIGIN_TO_REDUCER_Y_L = 0.7135 * u.cm

#Distance from the lever arm origin to the center of the reducer in the
# x direction.
LA_ORIGIN_TO_REDUCER_CENTER_X_L = 0.0290 * u.m

#Distance from the lever arm origin to the center of the reducer in the
# y direction.
LA_ORIGIN_TO_REDUCER_CENTER_Y_L = 0.7919 * u.m

LEVER_BRACKET_W = 0.625 * u.inch

LEVER_BRAKCET_L = 1.5 * u.inch

RADIUS_LA_BAR = 0.375 * u.inch

LEVER_BRACKET_THICKNESS = 0.08 * u.inch

DIAM_LA_BAR = 0.375 * u.inch

LA_BAR_L = 4 * u.inch

SLIDER_L = 3 * u.inch

SLIDER_W = 3.2 * 10**-3 * u.m

DROPTUBE_ND = 0.5 * u.inch

SLIDER_H = 0.625 * u.inch

DROPTUBE_L = 0.61 * u.m

#The length of the drop tube needs to be calculated. The drop tube must be
# as long as the supercritical flow.
#Thus the drop tube must extend down to the elevation of the sed tank
# effluent weir. This constant should be removed!

#Outer diameter of fitting- measured from CDC research team's fittin
OUTER_D_CDC_FITTING = 5/32 * u.inch

#Inner diameter of fitting- measured from CDC research team's fitting
CDC_FITTING_ID = 0.126 * u.inch

#Length of fitting - measured from CDC research team's fitting
CDC_FITTING_L = 0.75 * u.inch

#st587 addition
##Constant Head Tank Dimensions

#five gallons bucket dimensions for constant head tanks

DIAM_CHT = 10 * u.cm

CHT_H = 37/3 * u.cm

CHT_WALL_THICKNESS = 1/3 * u.cm

DELIVERY_PIPE_ND = 0.6 * u.inch

PIPE_SCHEDULE_FLEX_TUBE = 2

PVC_BALL_VALVE_L = 0.1625/4 * u.cm

MOUNTING_BOARD_THICKNESS = 1.5 * u.inch

##Manifold Dimensions

SPACE_CDC_LEVER_TO_MANIFOLD = 40 * u.cm

CHLOR_AIR_RELEASE_PIPE_L = 30 * u.cm #Arbitratily selected

# 0 is alum, 1 is PACl
# EN_COAG=1

COAG_SACK_MASS = 25 * u.kg

# The coagulant stock is relatively stable and can last many days. Here we
# set the minimum time the coagulant stock will last when applying the
# maximum possible dose to size the stock tanks. In general the dose will
# be less than this and the stock will last much longer.
COAG_STOCK_MIN_EST_TIME = 1 * u.day

# Want chlorine stock to run out on average every day so that the stock
# is made fresh frequently because the chlorine stock degrades with time
# depending on temperature, concentration, and pH.
CHLOR_STOCK_AVE_TIME = 1 * u.day

COAG_TUBE_ID = 0.125 * u.inch
# 1/8" tubes are readily available in hardware stores in Honduras
CHLOR_TUBE_ID = 0.125 * u.inch

COAG_STOCK_EST_CONC = 150 * u.g / u.L

CHLOR_STOCK_EST_CONC = 15 * u.g / u.L

CHLOR_P = 0.7

# This is the elevation difference between the outlet of the coagulant
# stock tanks and the water level in the constant head tank, which is set
# by the hydraulic head required to provide the desired max chemical flow
# rate through the float valve orifice in the CHT.
# It is treated as constant here to ensure a practical elevation difference
# is left between the stock tanks and the CHT even when a float valve is
# selected which requires very little hydraulic head to deliver the
# required maximum chemical flow rate.
COAG_TANK_ABOVE_HEAD_TANK_H = 30 * u.cm

# This is the distance from the bottom of the stock tanks to the outlets
# to allow space for solids to settle.
CENTER_STOCK_OUTLET_DIST = 10 * u.cm

# Distance between a tank and the border of the platform
CHEM_TANK_BORDER_S = 5 * u.cm

# This is the estimated elevation difference between the water level in
# the constant head tank and the top of the entrance tank wall.
# The constant head tank water level is the same as the elevation of the
# outlet of the dosing tube when the lever arm is horizontal (zero flow).
# Therefore this height depends only on the hardware used to make the
# slider/drop tube assembly and to mount the lever arm to the entrance
# tank wall.
# Note that this will vary depending on hardware used, and is only
# defined here to calculate the elevation of the stock tanks, which can
# be approximate.
DOSER_ASSEMBLY_H = 6.77 * u.cm

# Maximum error allowed between a linear flow vs tube head loss
# relationship and the actual performance (which is affected by non-linear
# minor losses), assuming calibration at the maximum flow rate.
LINEAR_CDC_ERROR_RATIO = 0.1

# Estimated minor loss coefficient for the small-diameter flexible tubing
# using fittings that have larger ID than the tubing.
CDC_TUBE_K_MINOR = 2

# Head loss through the doser at maximum flow rate.
# Maximum head loss through the small-diameter dosing tubing, which
# corresponds to the variation in water levels in the entrance tank and
# the difference between the maximum and minimum elevation of the dosing
# tube outlet attached to the lever arm.
HL = 20 * u.cm

# Estimated distance between fluid level in constant head tank and float
# valve orifice
FLOAT_VALVE_H = 5 * u.cm

# Nominal diameter of the PVC plumbing for the chlorine dosing system.
CHLOR_PIPE_ND = 0.5 * u.inch

# Nominal diameter of the PVC plumbing for the coagulant dosing system.
COAG_PIPE_ND = 0.5 * u.inch

# Supplier Information:
# http://www.rotoplas.com/assets/files/industria/catalogo.pdf
# 5-gallon bucket
# http://www.mcmaster.com/#storage-buckets/=kd23oh
# 35-gallon drum
# http://www.jlmovingsupplies.com/c31/DIXIE-OPEN-CLOSED-HEAD-DRUMS-p36721.html
VOL_CHEM_TANK_AVAIL = [5 * u.gal, 35 * u.gal, 55 * u.gal,
                       450 * u.L, 750 * u.L, 1100 * u.L, 2500 * u.L]

D_CHEM_TANK_AVAIL = [11.875 * u.inch, 20.75 * u.inch, 22.5 * u.inch,
                     0.85 * u.m, 1.10 * u.m, 1.10 * u.m, 1.55 * u.m]

CHEM_TANK_AVAIL_H = [17.75 * u.inch, 31.75 * u.inch, 33.5 * u.inch,
                     0.99 * u.inch, 1.02 * u.inch, 1.39 * u.inch,
                     1.65 * u.inch]

##added 12/5/16 by mrf222 ensures weir does not overtop backwards if
# filter weir is too high
SED_WEIR_FREE_BOARD_H = 5 * u.cm

# Space beyond the entrance tank in the plant drain channel where the
# drop pipes from the CDC lever arm can come down and be connected with
# the chlorine and coagulant dosing points.
CHEM_LEVER_ARM_SPACE_L = 75 * u.cm
DIAM_TUBE_ENGLISH = [1, 2, 3 ,4 ,5 ,6, 7]*u.inch/16
DIAM_TUBE_METRIC = [2, 3, 4, 6, 8, 10]*u.mm
DIAM_FLT_VLV_ORIFICES_AVAIL = [0.093, 0.187, 0.25, 0.312]*u.inch



from aguaclara.core import drills


def test_drill():
    print(
        "Imperial drill bit diameters: \n"
        + repr(drills.DRILL_BITS_D_IMPERIAL) + "\n"
    )
    print(
        "Metric drill bit diameters: \n"
        + repr(drills.DRILL_BITS_D_METRIC) + "\n"
    )



from aguaclara.core.units import unit_registry as u
from aguaclara.design.floc import Flocculator


class Plant:
    """Functions for designing an AguaClara water treatment plant."""

    def __init__(self, q=20 * u.L/u.s, temp=25 * u.degC):
        """Initialize a Plant object that represents a real AguaClara water
        treatment plant.

        :param q: Flow rate of water through the plant.
        :type q: float * u.L/u.s
        :param temp: Temperature of water through the plant.
        :type q: float * u.degC
        :returns: object
        :rtype: Plant
        """
        self.q = q
        self.temp = temp
        self.floc = Flocculator(Q=q, temp=temp)

    @property
    def ent_tank_a(self):
        """Calculate the planview area of the entrance tank, given the volume of
        the flocculator.

        :returns: The planview area of the entrance tank.
        :rtype: float * u.m ** 2
        """
        # first guess planview area
        a_new = 1 * u.m**2
        a_ratio = 2  # set to >1+tolerance to start while loop
        tolerance = 0.01
        a_floc_pv = (
            self.floc.vol /
            (self.floc.downstream_H + (self.floc.HL / 2))
        )
        while a_ratio > (1 + tolerance):
            a_et_pv = a_new
            a_etf_pv = a_et_pv + a_floc_pv
            w_tot = a_etf_pv / self.floc.max_L
            w_chan = w_tot / self.floc.channel_n

            a_new = self.floc.max_L * w_chan
            a_ratio = a_new / a_et_pv
        return a_new



# -*- coding: utf-8 -*-
"""
Created on Fri Jul 7 11:51:51 2017

@author: Sage Weber-Shirk

Last modified: Tue Jun 4 2019
By: Hannah Si
"""

#Note: All answer values in this file should be checked against MathCad
#before this file is released to the Master branch!

from aguaclara.core.units import unit_registry as u
import unittest

developing = False
if developing:
    import sys
    sys.path.append("../../aguaclara/core")
    import physchem as pc
else:
    from aguaclara.core import physchem as pc

class AirTest(unittest.TestCase):
    """Test the air density function"""
    def assertAlmostEqualQuantity(self, first, second, places=7):
        self.assertAlmostEqual(first.magnitude, second.magnitude, places)
        self.assertEqual(first.units, second.units, places)

    def test_air_density(self):
        answer = 1.29320776*u.kg/u.m**3
        self.assertAlmostEqualQuantity(pc.density_air(1*u.atm, 28.97*u.g/u.mol, 273*u.K), answer)
        answer = 2.06552493*u.kg/u.m**3
        self.assertAlmostEqualQuantity(pc.density_air(5*u.atm, 10*u.g/u.mol, 295*u.K), answer)
        answer = 1.62487961*u.kg/u.m**3
        self.assertAlmostEqualQuantity(pc.density_air(101325*u.Pa, 40*u.g/u.mol, 300*u.K), answer)
        answer = 0.20786109*u.kg/u.m**3
        self.assertAlmostEqualQuantity(pc.density_air(700*u.mmHg, 5*u.g/u.mol, 270*u.K), answer)
        answer = 0*u.kg/u.m**3
        self.assertAlmostEqualQuantity(pc.density_air(0*u.atm, 28.97*u.g/u.mol, 273*u.K), answer)

class GeometryTest(unittest.TestCase):
    """Test the circular area and diameter functions."""
    def test_area_circle(self):
        """area_circle should  should give known result with known input."""
        checks = ((1, 0.7853981633974483*u.m**2),
                  (495.6, 192908.99423885669*u.m**2))
        for i in checks:
            with self.subTest(i=i):
                self.assertAlmostEqual(pc.area_circle(i[0]), i[1])

    def test_area_circle_units(self):
        """area_circle should  should give known result with known input and correct units"""
        checks = ((1*u.m, 7853.981633974483*u.cm**2),
                  (495.6*u.cm, 19.290899423885669*u.m**2))
        for i in checks:
            with self.subTest(i=i):
                self.assertAlmostEqual(pc.area_circle(i[0]), i[1])

    def test_area_circle_range(self):
        """area_circle should return errors with inputs <= 0."""
        checks = (0, -3)
        for i in checks:
            with self.subTest(i=i):
                self.assertRaises(ValueError, pc.area_circle, i)

    def test_diam_circle(self):
        """diam_circle should should give known result with known input."""
        checks = ((1, 1.1283791670955126),
                  (0.1, 0.3568248232305542),
                  (347, 21.019374919894773),
                  (10000 * u.cm**2, 1.1283791670955126))
        for i in checks:
            with self.subTest(i=i):
                self.assertAlmostEqual(pc.diam_circle(i[0]).magnitude, i[1])

    def test_diam_circle_range(self):
        """diam_circle should return errors with inputs <= 0."""
        checks = ((0, ValueError),
                  (-3, ValueError))
        for i in checks:
            with self.subTest(i=i):
                self.assertRaises(i[1], pc.diam_circle, i[0])


class WaterPropertiesTest(unittest.TestCase):
    """Test the density and dynamic/kinematic viscosity functions."""
    def test_water_table(self):
        """The table density_water relies upon shouldn't need to be changed."""
        table = pc.WATER_DENSITY_TABLE
        checks = ((0, 273.15, 999.9),
                  (4, 303.15, 995.7),
                  (11, 373.15, 958.4))
        for i in checks:
            with self.subTest(i=i):
                self.assertAlmostEqual(table[0][i[0]], i[1])
                self.assertAlmostEqual(table[1][i[0]], i[2])

    def test_water_table_units(self):
        """The water density table should handle units properly."""
        table = pc.WATER_DENSITY_TABLE
        self.assertAlmostEqual(table[0][0], (0 * u.degC).to_base_units().magnitude)
        self.assertAlmostEqual(table[0][4], (30 * u.degC).to_base_units().magnitude)

    def test_density_water_true(self):
        """density_water should give known result with known input."""
        checks = ((273.15, 999.9),
                  (300, 996.601907542082),
                  (343.15, 977.8))
        for i in checks:
            with self.subTest(i=i):
                self.assertAlmostEqual(pc.density_water(i[0]).magnitude, i[1])

    def test_viscosity_dynamic(self):
        """viscosity_dynamic should give known result with known input."""
        checks = ((300, 0.0008540578046518858),
                  (372, 0.00028238440851243975),
                  (274, 0.0017060470223965783))
        for i in checks:
            with self.subTest(i=i):
                self.assertAlmostEqual(pc.viscosity_dynamic(i[0]).magnitude, i[1])

    def test_viscosity_dynamic_units(self):
        """viscosity_dynamic should give known result with known input."""
        checks = ((300 * u.degK, 0.0008540578046518858),
                  (26.85 * u.degC, 0.0008540578046518858))
        for i in checks:
            with self.subTest(i=i):
                self.assertAlmostEqual(pc.viscosity_dynamic(i[0]).magnitude, i[1])

    def test_viscosity_kinematic(self):
        """nu should give known results with known input."""
        checks = ((342, 4.1584506710898959e-07),
                  (297, 9.1670473903811879e-07),
                  (273.15, 1.7532330683680798e-06),
                  (373.15, 2.9108883329847625e-07))
        for i in checks:
            with self.subTest(i=i):
                self.assertAlmostEqual(pc.viscosity_kinematic(i[0]).magnitude, i[1])

    def test_viscosity_kinematic_units(self):
        """nu should handle units correctly."""
        checks = ((342, 4.1584506710898959e-07),
                  (297 * u.degK, 9.1670473903811879e-07),
                  (0 * u.degC, 1.7532330683680798e-06),
                  (100 * u.degC, 2.9108883329847625e-07))
        for i in checks:
            with self.subTest(i=i):
                self.assertAlmostEqual(pc.viscosity_kinematic(i[0]).magnitude, i[1])
                self.assertAlmostEqual(pc.viscosity_kinematic(i[0]),
                                       (pc.viscosity_dynamic(i[0]) / pc.density_water(i[0])))


class ReynoldsNumsTest(unittest.TestCase):
    """Test the various Reynolds Number functions."""
    def test_re_pipe(self):
        """re_pipe should return known results with known input."""
        checks = (((12, 6, 0.01), 254.64790894703253),
                  ((60, 1, 1), 76.39437268410977),
                  ((1, 12, .45), 0.23578510087688198))
        for i in checks:
            with self.subTest(i=i):
                self.assertAlmostEqual(pc.re_pipe(*i[0]), i[1])

    def test_re_pipe_range(self):
        """re_pipe should raise errors when inputs are out of bounds."""
        checks = ((0, 4, .5), (1, 0, .4), (1, 1, -0.1), (1, 1, 0))
        for i in checks:
            with self.subTest(i=i):
                self.assertRaises(ValueError, pc.re_pipe, *i)

    def test_re_pipe_units(self):
        """re_pipe should handle units correctly."""
        base = pc.re_pipe(12, 6, 0.01)
        checks = ([12 * u.m**3/u.s, 6 * u.m, 0.01 * u.m**2/u.s],
                  [12000 * u.L/u.s, 600 * u.cm, 0.000001 * u.ha/u.s],
                  [12, 0.006 * u.km, 100 * u.cm**2/u.s])
        for i in checks:
            with self.subTest(i=i):
                self.assertAlmostEqual(pc.re_pipe(*i), base)

    def test_re_rect(self):
        """re_rect should return known result with known input."""
        self.assertAlmostEqual(pc.re_rect(10, 4, 6, 1, True), 2.5)

    def test_re_rect_range(self):
        """re_rect should raise errors when inputs are out of bounds."""
        checks = ((0, 1, 1, 1, False), (1, 1, 1, 0, False))
        for i in checks:
            with self.subTest(i=i):
                self.assertRaises(ValueError, pc.re_rect, *i)

    def test_re_rect_units(self):
        """re_rect should handle units correctly."""
        base = pc.re_rect(10, 4, 6, 1, True)
        checks = ([10 * u.m**3/u.s, 4 * u.m, 6 * u.m, 1 * u.m**2/u.s, True],
                  [10000 * u.L/u.s, 4, 6, 1, True],
                  [10, 400 * u.cm, 6, 1, True],
                  [10, 4, 0.006 * u.km, 1, True],
                  [10, 4, 6, 0.0001 * u.ha/u.s, True])
        for i in checks:
            with self.subTest(i=i):
                self.assertAlmostEqual(pc.re_rect(*i), base)

    def test_re_general(self):
        """re_general should return known values with known input."""
        checks = (([1, 2, 3, 0.4], 6.666666666666666),
                  ([17, 6, 42, 1], 9.714285714285714),
                  ([0, 1, 2, 0.3], 0))
        for i in checks:
            with self.subTest(i=i):
                self.assertAlmostEqual(pc.re_general(*i[0]), i[1])

    def test_re_general_range(self):
        """re_general should raise errors when inputs are out of bounds."""
        checks = ((-1, 2, 3, 0.4), (1, 2, 3, 0))
        for i in checks:
            with self.subTest(i=i):
                self.assertRaises(ValueError, pc.re_general, *i)

    def test_re_general_units(self):
        """re_general should handle units correctly."""
        base = pc.re_general(1, 2, 3, 0.4)
        checks = ([1 * u.m/u.s, 2 * u.m**2, 3 * u.m, 0.4 * u.m**2/u.s],
                  [100 * u.cm/u.s, 2, 3, 0.4],
                  [1, 20000 * u.cm**2, 3, 0.4],
                  [1, 2, 0.003 * u.km, 0.4],
                  [1, 2, 3, 4000 * u.cm**2/u.s])
        for i in checks:
            with self.subTest(i=i):
                self.assertAlmostEqual(pc.re_general(*i), base)


class RadiusFuncsTest(unittest.TestCase):
    """Test the various radius-acquisition functions."""
    def test_radius_hydraulic(self):
        """radius_hydraulic should return known results with known input."""
        checks = (([10, 4, False], 1.4285714285714286),
                  ([10, 4, True], 2.2222222222222223))
        for i in checks:
            with self.subTest(i=i):
                self.assertAlmostEqual(pc.radius_hydraulic(*i[0]).magnitude, i[1])

    def test_radius_hydraulic_range(self):
        """radius_hydraulic should raise errors when inputs are out of bounds."""
        checks = (([0, 4, True], ValueError), ([-1, 4, True], ValueError),
                  ([1, 0, True], ValueError), ([10, -1, True], ValueError),
                  ([10, 4, 0], TypeError), ([10, 4, 6], TypeError))
        for i in checks:
            with self.subTest(i=i):
                self.assertRaises(i[1], pc.radius_hydraulic, *i[0])

    def test_radius_hydraulic_units(self):
        """radius_hydraulic should handle units correctly."""
        base = pc.radius_hydraulic(10, 4, False)
        checks = ([1000 * u.cm, 4, False],
                  [0.01 * u.km, 40 * u.dm, False])
        for i in checks:
            with self.subTest(i=i):
                self.assertAlmostEqual(pc.radius_hydraulic(*i), base)

    def test_radius_hydraulic_general(self):
        """radius_hydraulic_general should return known results with known input."""
        checks = (([6, 12], 0.5), ([70, 0.4], 175))
        for i in checks:
            with self.subTest(i=i):
                self.assertAlmostEqual(pc.radius_hydraulic_general(*i[0]).magnitude, i[1])

    def test_radius_hydraulic_general_range(self):
        """radius_hydraulic_general should not accept inputs of 0 or less."""
        checks = ([0, 6], [6, 0])
        for i in checks:
            with self.subTest(i=i):
                self.assertRaises(ValueError, pc.radius_hydraulic_general, *i)

    def test_radius_hydraulic_general_units(self):
        """radius_hydraulic_general should handle units correctly."""
        base = pc.radius_hydraulic_general(4, 7)
        checks = ([4 * u.m**2, 7 * u.m], [40000 * u.cm**2, 7],
                  [4, 0.007 * u.km])
        for i in checks:
            with self.subTest(i=i):
                self.assertAlmostEqual(pc.radius_hydraulic_general(*i), base)


class FrictionFuncsTest(unittest.TestCase):
    """Test the friction functions."""
    def test_fric(self):
        """fric should return known results with known input."""
        checks = (([100, 2, 0.001, 1], 0.33154589118654193),
                  ([100, 2, 0.1, 1], 0.10053096491487337),
                  ([100, 2, 0.001, 0], 0.019675384283293733),
                  ([46, 9, 0.001, 0.03], 0.039382681891291252),
                  ([55, 0.4, 0.5, 0.0001], 0.18278357257249706))
        for i in checks:
            with self.subTest(i=i):
                self.assertAlmostEqual(pc.fric(*i[0]), i[1])

    def test_fric_range(self):
        """fric should raise an error if 0 <= PipeRough <= 1 is not true."""
        checks = ([1, 2, 0.1, -0.1],
                  [1, 2, 0.1, 1.1])
        for i in checks:
            with self.subTest(i=i):
                self.assertRaises(ValueError, pc.fric, *i)

    def test_fric_units(self):
        """fric should handle units correctly."""
        base = pc.fric(100, 2, 0.001, 1)
        checks = ([100 * u.m**3/u.s, 2 * u.m, 0.001 * u.m**2/u.s, 1 * u.m],
                  [100000 * u.L/u.s, 2, 0.001, 1],
                  [100, 20 * u.dm, 0.001, 1],
                  [100, 2, 10 * u.cm**2/u.s, 1],
                  [100, 2, 0.001, 1000 * u.mm])
        for i in checks:
            with self.subTest(i=i):
                self.assertAlmostEqual(pc.fric(*i), base)

    def test_fric_rect(self):
        """fric_rect should return known results with known inputs."""
        checks = (([60, 0.7, 1, 0.6, 0.001, True], 0.432),
                  ([60, 0.7, 1, 0.6, 0.001, False], 0.544),
                  ([120, 1, 0.04, 0.125, 0.6, True], 150.90859874356411),
                  ([120, 1, 0.04, 0.125, 0.6, False], 0.034666666666666665),
                  ([120, 1, 0.04, 0.125, 0, False], 0.034666666666666665),
                  ([120, 1, 0.04, 0.125, 0, True], 0.042098136441473824))
        for i in checks:
            with self.subTest(i=i):
                self.assertAlmostEqual(pc.fric_rect(*i[0]), i[1])

    def test_fric_rect_range(self):
        """fric_rect should raise an error if 0 <= PipeRough <= 1 is not true."""
        checks = ([1, 1, 1, 1, 1.1, True],)
        for i in checks:
            with self.subTest(i=i):
                self.assertRaises(ValueError, pc.fric_rect, *i)

    def test_fric_rect_units(self):
        """fric_rect should handle units correctly."""
        base = pc.fric_rect(0.06, 0.1, 0.0625, 0.347, 0.06, True)
        checks = ([0.06 * u.m**3/u.s, 0.1 * u.m, 0.0625 * u.m,
                   0.347 * u.m**2/u.s, 0.06 * u.m, True],
                  [60 * u.L/u.s, 0.1, 0.0625, 0.347, 0.06, True],
                  [0.06, 10 * u.cm, 0.0625, 0.347, 0.06, True],
                  [0.06, 0.1, 6.25 * u.cm, 0.347, 0.06, True],
                  [0.06, 0.1, 0.0625, 3470 * u.cm**2/u.s, 0.06, True],
                  [0.06, 0.1, 0.0625, 0.347, 6 * u.cm, True])
        for i in checks:
            with self.subTest(i=i):
                self.assertAlmostEqual(pc.fric_rect(*i), base)

    def test_fric_general(self):
        """fric_general should return known results with known inputs."""
        checks = (([9, 0.67, 3, 0.987, 0.86], 0.3918755555555556),
                  ([1, 1, 1, 1, 1], 16),
                  ([120, 0.6, 12, 0.3, 0.002], 0.023024557179148988))
        for i in checks:
            with self.subTest(i=i):
                self.assertAlmostEqual(pc.fric_general(*i[0]), i[1])

    def test_fric_general_range(self):
        """fric_general should raise an error if 0 <= PipeRough <= 1 is not true."""
        checks = ((1, 1, 1, 1, -0.0001), (1, 1, 1, 1, 1.1))
        for i in checks:
            with self.subTest(i=i):
                self.assertRaises(ValueError, pc.fric_general, *i)

    def test_fric_general_units(self):
        """fric_general should handle units correctly."""
        base = pc.fric_general(46.2, 0.75, 1.23, 0.46, 0.002)
        checks = ([46.2 * u.m**2, 0.75 * u.m, 1.23 * u.m/u.s,
                   0.46 * u.m**2/u.s, 0.002 * u.m],
                  [462000 * u.cm**2, 0.75, 1.23, 0.46, 0.002],
                  [46.2, 750 * u.mm, 1.23, 0.46, 0.002],
                  [46.2, 0.75, 0.00123 * u.km/u.s, 0.46, 0.002],
                  [46.2, 0.75, 1.23, 4600 * u.cm**2/u.s, 0.002],
                  [46.2, 0.75, 1.23, 0.46, 2 * u.mm])
        for i in checks:
            with self.subTest(i=i):
                self.assertAlmostEqual(pc.fric_general(*i), base)


class HeadlossFuncsTest(unittest.TestCase):
    """Test the headloss functions."""
    def test_headloss_fric(self):
        """headloss_fric should return known results with known inputs."""
        checks = (([100, 2, 4, 0.001, 1], 34.2549414191127),
                  ([100, 2, 4, 0.1, 1], 10.386744054168654),
                  ([100, 2, 4, 0.001, 0], 2.032838149828097),
                  ([46, 9, 12, 0.001, 0.03], 0.001399778168304583),
                  ([55, 0.4, 2, 0.5, 0.0001], 8926.108171551185))
        for i in checks:
            with self.subTest(i=i):
                self.assertAlmostEqual(pc.headloss_fric(*i[0]).magnitude, i[1])

    def test_headloss_fric_range(self):
        """headloss_fric should raise an error if Length <= 0."""
        checks = ([1, 1, 0, 1, 1], [1, 1, -1, 1, 1])
        for i in checks:
            with self.subTest(i=i):
                self.assertRaises(ValueError, pc.headloss_fric, *i)

    def test_headloss_fric_units(self):
        """headloss_fric should handle units correctly."""
        base = pc.headloss_fric(100, 2, 4, 0.001, 0.03)
        checks = ([100 * u.m**3/u.s, 2 * u.m, 4 * u.m,
                   0.001 * u.m**2/u.s, 0.03 * u.m],
                  [10**5 * u.L/u.s, 2, 4, 0.001, 0.03],
                  [100, 200 * u.cm, 4, 0.001, 0.03],
                  [100, 2, 4000 * u.mm, 0.001, 0.03],
                  [100, 2, 4, 10 * u.cm**2/u.s, 0.03],
                  [100, 2, 4, 0.001, 3 * u.cm])
        for i in checks:
            with self.subTest(i=i):
                self.assertAlmostEqual(pc.headloss_fric(*i), base)

    def test_headloss_exp(self):
        """headloss_exp should return known results with known input."""
        self.assertAlmostEqual(pc.headloss_exp(60, 0.9, 0.067).magnitude,
                         30.386230766265214)

    def test_headloss_exp_range(self):
        """headloss_exp should raise errors when inputs are out of bounds."""
        checks = ([0, 1, 1], [1, 0, 1], [1, 1, -1])
        for i in checks:
            with self.subTest(i=i):
                self.assertRaises(ValueError, pc.headloss_exp, *i)
        pc.headloss_exp(1, 1, 0)

    def test_headloss_exp_units(self):
        """headloss_exp should handle units correctly."""
        base = pc.headloss_exp(60, 0.9, 0.067).magnitude
        checks = ([60 * u.m**3/u.s, 0.9 * u.m, 0.067 * u.dimensionless],
                  [60000 * u.L/u.s, 0.9, 0.067],
                  [60, 900 * u.mm, 0.067])
        for i in checks:
            with self.subTest(i=i):
                self.assertAlmostEqual(pc.headloss_exp(*i).magnitude, base)

    def test_headloss(self):
        """headloss should return known results with known inputs."""
        checks = (([100, 2, 4, 0.001, 1, 2], 137.57379509731857),
                  ([100, 2, 4, 0.1, 1, 0.4], 31.05051478980984),
                  ([100, 2, 4, 0.001, 0, 1.2], 64.024150356751633),
                  ([46, 9, 12, 0.001, 0.03, 4], 0.10802874052554703),
                  ([55, 0.4, 2, 0.5, 0.0001, 0.12], 10098.131417963332))
        for i in checks:
            with self.subTest(i=i):
                self.assertAlmostEqual(pc.headloss(*i[0]).magnitude, i[1])

    def test_headloss_units(self):
        """headloss should handle units correctly."""
        base = pc.headloss(100, 2, 4, 0.001, 1, 2)
        checks = ([100 * u.m**3/u.s, 2 * u.m, 4 * u.m,
                   0.001 * u.m**2/u.s, 1 * u.m, 2],
                  [10**5 * u.L/u.s, 2, 4, 0.001, 1, 2],
                  [100, 200 * u.cm, 4, 0.001, 1, 2],
                  [100, 2, 4000 * u.mm, 0.001, 1, 2],
                  [100, 2, 4, 10 * u.cm**2/u.s, 1, 2],
                  [100, 2, 4, 0.001, 100 * u.cm, 2])
        for i in checks:
            with self.subTest(i=i):
                self.assertAlmostEqual(pc.headloss(*i), base)

    def test_headloss_fric_rect(self):
        """headloss_fric_rect should return known result with known inputs."""
        checks = (([0.06, 3, 0.2, 4, 0.5, 0.006, True], 1.3097688246694272),
                  ([0.06, 3, 0.2, 4, 0.5, 0.006, False], 4.640841787063992))
        for i in checks:
            with self.subTest(i=i):
                self.assertAlmostEqual(pc.headloss_fric_rect(*i[0]).magnitude, i[1])

    def test_headloss_fric_rect_range(self):
        """headloss_fric_rect should raise an error when Length <=0."""
        checks = ((1, 1, 1, 0, 1, 1, 1), (1, 1, 1, -1, 1, 1, 1))
        for i in checks:
            with self.subTest(i=i):
                self.assertRaises(ValueError, pc.headloss_fric_rect, *i)

    def test_headloss_fric_rect_units(self):
        """headloss_fric_rect should handle units correctly."""
        base = pc.headloss_fric_rect(0.06, 2, 0.004, 3, 0.89, 0.07, True)
        checks = ([0.06 * u.m**3/u.s, 2 * u.m, 0.004 * u.m, 3 * u.m,
                   0.89 * u.m**2/u.s, 0.07 * u.m, True],
                  [60 * u.L/u.s, 2, 0.004, 3, 0.89, 0.07, True],
                  [0.06, 200 * u.cm, 0.004, 3, 0.89, 0.07, True],
                  [0.06, 2, 4 * u.mm, 3, 0.89, 0.07, True],
                  [0.06, 2, 0.004, 300 * u.cm, 0.89, 0.07, True],
                  [0.06, 2, 0.004, 3, 8900 * u.cm**2/u.s, 0.07, True],
                  [0.06, 2, 0.004, 3, 0.89, 7 * u.cm, True])
        for i in checks:
            with self.subTest(i=i):
                self.assertAlmostEqual(pc.headloss_fric_rect(*i), base)

    def test_headloss_exp_rect(self):
        """headloss_exp_rect should return known result for known input."""
        checks = ([0.06, 2, 0.004, 1], 2.8679518490004234)
        self.assertAlmostEqual(pc.headloss_exp_rect(*checks[0]).magnitude, checks[1])

    def test_headloss_exp_rect_range(self):
        """headloss_exp_rect should raise errors when inputs are out of bounds."""
        checks = ((0, 1, 1, 1), (1, 0, 1, 1), (1, 1, 0, 1), (1, 1, 1, -1))
        for i in checks:
            with self.subTest(i=i):
                self.assertRaises(ValueError, pc.headloss_exp_rect, *i)
        pc.headloss_exp_rect(1, 1, 1, 0)

    def test_headloss_exp_rect_units(self):
        """headloss_exp_rect should handle units correctly."""
        base = pc.headloss_exp_rect(0.06, 2, 0.9, 1)
        checks = ([0.06 * u.m**3/u.s, 2 * u.m, 0.9 * u.m, 1 * u.dimensionless],
                  [60 * u.L/u.s, 2, 0.9, 1],
                  [0.06, 200 * u.cm, 0.9, 1],
                  [0.06, 2, 900 * u.mm, 1])
        for i in checks:
            with self.subTest(i=i):
                self.assertAlmostEqual(pc.headloss_exp_rect(*i), base)

    def test_headloss_rect(self):
        """headloss_rect should return known result for known inputs."""
        checks = (([0.06, 3, 0.2, 4, 1, 0.5, 0.006, True], 1.3102786827759163),
                  ([0.06, 3, 0.2, 4, 1, 0.5, 0.006, False], 4.641351645170481))
        for i in checks:
            with self.subTest(i=i):
                self.assertAlmostEqual(pc.headloss_rect(*i[0]).magnitude, i[1])

    def test_headloss_rect_units(self):
        """headloss_rect should handle units properly."""
        base = pc.headloss_rect(0.06, 3, 0.2, 4, 1, 0.5, 0.006, True)
        checks = ([0.06 * u.m**3/u.s, 3 * u.m, 0.2 * u.m, 4 * u.m,
                   1, 0.5 * u.m**2/u.s, 0.006 * u.m, True],
                  [60 * u.L/u.s, 3, 0.2, 4, 1, 0.5, 0.006, True],
                  [0.06, 300 * u.cm, 0.2, 4, 1, 0.5, 0.006, True],
                  [0.06, 3, 200 * u.mm, 4, 1, 0.5, 0.006, True],
                  [0.06, 3, 0.2, 0.004 * u.km, 1, 0.5, 0.006, True],
                  [0.06, 3, 0.2, 4, 1, 5000 * u.cm**2/u.s, 0.006, True],
                  [0.06, 3, 0.2, 4, 1, 0.5, 6 * u.mm, True])
        for i in checks:
            with self.subTest(i=i):
                self.assertAlmostEqual(pc.headloss_rect(*i), base)
        self.assertRaises(ValueError, pc.headloss_rect,
                          *[1, 1, 1, 1, 1 * u.m, 1, 1, True])

    def test_headloss_fric_general(self):
        """headloss_fric_general should return known result for known inputs."""
        checks = (([1, 1, 1, 1, 1, 1], 0.20394324259558566),
                  ([25, 4, 0.6, 2, 1, 1], 0.006265136412536391))
        for i in checks:
            with self.subTest(i=i):
                self.assertAlmostEqual(pc.headloss_fric_general(*i[0]).magnitude, i[1])

    def test_headloss_fric_general_range(self):
        """headloss_fric_general should raise an error when Length <= 0."""
        checks = ([1, 1, 1, 0, 1, 1], [1, 1, 1, -1, 1, 1])
        for i in checks:
            with self.subTest(i=i):
                self.assertRaises(ValueError, pc.headloss_fric_general, *i)

    def test_headloss_fric_general_units(self):
        """headloss_fric_general should handle units correctly."""
        base = pc.headloss_fric_general(36, 5, 0.2, 6, 0.4, 0.002)
        checks = ([36 * u.m**2, 5 * u.m, 0.2 * u.m/u.s,
                   6 * u.m, 0.4 * u.m**2/u.s, 0.002 * u.m],
                  [0.000036 * u.km**2, 5, 0.2, 6, 0.4, 0.002],
                  [36, 500 * u.cm, 0.2, 6, 0.4, 0.002],
                  [36, 5, 20 * u.cm/u.s, 6, 0.4, 0.002],
                  [36, 5, 0.2, 0.006 * u.km, 0.4, 0.002],
                  [36, 5, 0.2, 6, 4000 * u.cm**2/u.s, 0.002],
                  [36, 5, 0.2, 6, 0.4, 2 * u.mm])
        for i in checks:
            with self.subTest(i=i):
                self.assertAlmostEqual(pc.headloss_fric_general(*i), base)

    def test_headloss_exp_general(self):
        """headloss_exp_general should return known result for known input."""
        self.assertAlmostEqual(pc.headloss_exp_general(0.06, 0.02).magnitude,
                         3.670978366720542e-06)

    def test_headloss_exp_general_range(self):
        """headloss_exp_general should raise errors if inputs are out of bounds."""
        checks = ((0,1), (1, -1))
        for i in checks:
            with self.subTest(i=i):
                self.assertRaises(ValueError, pc.headloss_exp_general, *i)
        pc.headloss_exp_general(1, 0)

    def test_headloss_exp_general_units(self):
        """headloss_exp_general should handle units correctly."""
        base = pc.headloss_exp_general(0.06, 0.02).magnitude
        checks = ([0.06 * u.m/u.s, 0.02 * u.dimensionless],
                  [6 * u.cm/u.s, 0.02])
        for i in checks:
            with self.subTest(i=i):
                self.assertAlmostEqual(pc.headloss_exp_general(*i).magnitude, base)

    def test_headloss_gen(self):
        """headloss_gen should return known value for known inputs."""
        checks = (([36, 0.1, 4, 6, 0.02, 0.86, 0.0045], 0.0013093911519979546),
                  ([49, 2.4, 12, 3, 2, 4, 0.6], 0.9396236839032805))
        for i in checks:
            with self.subTest(i=i):
                self.assertAlmostEqual(pc.headloss_gen(*i[0]).magnitude, i[1])

    def test_headloss_gen_units(self):
        """headloss_gen should handle units correctly."""
        base = pc.headloss_gen(49, 2.4, 12, 3, 2, 4, 0.6).magnitude
        checks = ([49 * u.m**2, 2.4 * u.m/u.s, 12 * u.m, 3 * u.m,
                   2 * u.dimensionless, 4 * u.m**2/u.s, 0.6 * u.m],
                  [490000 * u.cm**2, 2.4, 12, 3, 2, 4, 0.6],
                  [49, 240 * u.cm/u.s, 12, 3, 2, 4, 0.6],
                  [49, 2.4, 0.012 * u.km, 3, 2, 4, 0.6],
                  [49, 2.4, 12, 3000 * u.mm, 2, 4, 0.6],
                  [49, 2.4, 12, 3, 2, 40000 * u.cm**2/u.s, 0.6],
                  [49, 2.4, 12, 3, 2, 4, 60 * u.cm])
        for i in checks:
            with self.subTest(i=i):
                self.assertAlmostEqual(pc.headloss_gen(*i).magnitude, base)

    def test_headloss_manifold(self):
        """headloss_manifold should return known value for known input."""
        checks = (([0.12, 0.4, 6, 0.8, 0.75, 0.0003, 5], 38.57715300752375),
                  ([2, 6, 40, 5, 1.1, 0.04, 6], 0.11938889890999548))
        for i in checks:
            with self.subTest(i=i):
                self.assertAlmostEqual(pc.headloss_manifold(*i[0]).magnitude, i[1])

    def test_headloss_manifold_range(self):
        """headloss_manifold should object if NumOutlets is not a positive int."""
        failChecks = ((1, 1, 1, 1, 1, 1, -1), (1, 1, 1, 1, 1, 1, 0),
                      (1, 1, 1, 1, 1, 1, 0.1))
        passchecks = ((1, 1, 1, 1, 1, 1, 1.0), (1, 1, 1, 1, 1, 1, 47))
        for i in failChecks:
            with self.subTest(i=i):
                self.assertRaises((ValueError, TypeError), pc.headloss_manifold, *i)
        for i in passchecks:
            with self.subTest(i=i):
                pc.headloss_manifold(*i)

    def test_headloss_manifold_units(self):
        """headloss_manifold should handle units correctly."""
        base = pc.headloss_manifold(2, 6, 40, 5, 1.1, 0.04, 6).magnitude
        unitchecks = ([2 * u.m**3/u.s, 6 * u.m, 40 * u.m, 5* u.dimensionless,
                       1.1 * u.m**2/u.s, 0.04 * u.m, 6* u.dimensionless],
                      [2 * u.m**3/u.s, 6, 40, 5, 1.1, 0.04, 6],
                      [2, 6000 * u.mm, 40, 5, 1.1, 0.04, 6],
                      [2, 6, 0.04 * u.km, 5, 1.1, 0.04, 6],
                      [2, 6, 40, 5, 11000 * u.cm**2/u.s, 0.04, 6],
                      [2, 6, 40, 5, 1.1, 4 * u.cm, 6])
        for i in unitchecks:
            with self.subTest(i=i):
                self.assertAlmostEqual(pc.headloss_manifold(*i).magnitude, base)


class OrificeFuncsTest(unittest.TestCase):
    """Test the orifice functions."""
    def test_flow_orifice(self):
        """flow_orifice should return known result for known input."""
        checks = (([0.4, 2, 0.46], 0.36204122788069698),
                  ([2, 0.04, 0.2], 0.55652566805118475),
                  ([7, 0, 1], 0),
                  ([1.4, 0.1, 0], 0))
        for i in checks:
            with self.subTest(i=i):
                self.assertAlmostEqual(pc.flow_orifice(*i[0]).magnitude, i[1])

    def test_flow_orifice_range(self):
        """flow_orifice should raise errors when inputs are out of bounds."""
        checks = ((0,1,1), (1, 1, 1.1), (1, 1, -1))
        for i in checks:
            with self.subTest(i=i):
                self.assertRaises(ValueError, pc.flow_orifice, *i)
        pc.flow_orifice(1, 1, 0)

    def test_flow_orifice_units(self):
        """flow_orifice should handle units correctly."""
        base = pc.flow_orifice(2, 3, 0.5).magnitude
        checks = ((2 * u.m, 3 * u.m, 0.5 * u.dimensionless),
                  (200 * u.cm, 3, 0.5),
                  (2, 0.003 * u.km, 0.5))
        for i in checks:
            with self.subTest(i=i):
                self.assertAlmostEqual(pc.flow_orifice(*i).magnitude, base)

    def test_flow_orifice_vert(self):
        """flow_orifice_vert should return known values for known inputs."""
        checks = (([1, 3, 0.4], 2.4077258053173911),
                  ([0.3, 4, 0.67], 0.41946278400781861), ([2, -4, 0.2], 0))
        for i in checks:
            with self.subTest(i=i):
                self.assertAlmostEqual(pc.flow_orifice_vert(*i[0]).magnitude, i[1])

    def test_flow_orifice_vert_range(self):
        """flow_orifice_vert should raise errors when inputs are out of bounds."""
        errorChecks = ((1, 1, -1), (1, 1, 2))
        errorlessChecks = ((1, 1, 1), (1, 1, 0), (1, 1, 0.5))
        for i in errorChecks:
            with self.subTest(i=i):
                self.assertRaises(ValueError, pc.flow_orifice_vert, *i)
        for i in errorlessChecks:
            with self.subTest(i=i):
                pc.flow_orifice_vert(*i)

    def test_flow_orifice_vert_units(self):
        """flow_orifice_vert should handle units correctly."""
        base = pc.flow_orifice_vert(1, 3, 0.4).magnitude
        checks = ([1 * u.m, 3 * u.m, 0.4 * u.dimensionless],
                  [100 * u.cm, 3, 0.4],
                  [1, 0.003 * u.km, 0.4])
        for i in checks:
            with self.subTest(i=i):
                self.assertAlmostEqual(pc.flow_orifice_vert(*i).magnitude, base)

    def test_head_orifice(self):
        """head_orifice should return known value for known inputs."""
        checks = (([1, 1, 1], 0.08265508294256473),
                  ([1.2, 0.1, 0.12], 0.05739936315455882),
                  ([2, 0.5, 0.04], 3.3062033177025895e-05))
        for i in checks:
            with self.subTest(i=i):
                self.assertAlmostEqual(pc.head_orifice(*i[0]).magnitude, i[1])

    def test_head_orifice_range(self):
        """head_orifice should raise errors when passed invalid inputs."""
        failChecks = ((0,1,1), (1, 1, 0), (1, -1, 1), (1, 2, 1))
        for i in failChecks:
            with self.subTest(i=i):
                self.assertRaises(ValueError, pc.head_orifice, *i)
        pc.head_orifice(1, 1, 1)
        self.assertRaises(ZeroDivisionError, pc.head_orifice, *(1, 0, 1))

    def test_head_orifice_units(self):
        """head_orifice should handle units correctly."""
        base = pc.head_orifice(2, 0.5, 0.04).magnitude
        checks = ([2 * u.m, 0.5 * u.dimensionless, 0.04 * u.m**3/u.s],
                  [200 * u.cm, 0.5, 0.04],
                  [2, 0.5, 40 * u.L/u.s])
        for i in checks:
            with self.subTest(i=i):
                self.assertAlmostEqual(pc.head_orifice(*i).magnitude, base)

    def test_area_orifice(self):
        """area_orifice should return known value for known inputs."""
        checks = (([3, 0.4, 0.06], 0.019554886342464974),
                  ([2, 0.1, 0.1], 0.15966497839052934),
                  ([0.5, 0.02, 3], 47.899493517158803))
        for i in checks:
            with self.subTest(i=i):
                self.assertAlmostEqual(pc.area_orifice(*i[0]).magnitude, i[1])

    def test_area_orifice_range(self):
        """area_orifice should raise errors when inputs are out of bounds."""
        failChecks = ((0, 1, 1), (1, 1, 0), (1, -1, 1), (1, 2, 1), (1, 0, 1))
        for i in failChecks:
            with self.subTest(i=i):
                self.assertRaises(ValueError, pc.area_orifice, *i)
        pc.area_orifice(1, 1, 1)

    def test_area_orifice_units(self):
        """area_orifice should handle units correctly."""
        base = pc.area_orifice(3, 0.4, 0.06).magnitude
        checks = ([3 * u.m, 0.4 * u.dimensionless, 0.06 * u.m**3/u.s],
                  [300 * u.cm, 0.4, 0.06],
                  [3, 0.4, 60 * u.L/u.s])
        for i in checks:
            with self.subTest(i=i):
                self.assertAlmostEqual(pc.area_orifice(*i).magnitude, base)

    def test_num_orifices(self):
        """num_orifices should return known value for known inputs."""
        checks = (([0.12, 0.04, 0.05, 2], 1),
                  ([6, 0.8, 0.08, 1.2], 6))
        for i in checks:
            with self.subTest(i=i):
                self.assertAlmostEqual(pc.num_orifices(*i[0]), i[1])

    def test_num_orifices_units(self):
        """num_orifices should handle units correctly."""
        base = pc.num_orifices(6, 0.8, 0.08, 1.2)
        checks = ([6 * u.m**3/u.s, 0.8 * u.dimensionless, 0.08 * u.m, 1.2 * u.m],
                  [6000 * u.L/u.s, 0.8, 0.08, 1.2],
                  [6, 0.8, 8 * u.cm, 1.2],
                  [6, 0.8, 0.08, 0.0012 * u.km])
        for i in checks:
            with self.subTest(i=i):
                self.assertAlmostEqual(pc.num_orifices(*i), base)

class FlowFuncsTest(unittest.TestCase):
    """Test the flow functions."""
    def test_flow_transition(self):
        """flow_transition should return known value for known inputs."""
        checks = (([2, 0.4], 1319.4689145077132),
                  ([0.8, 1.1], 1451.4158059584847))
        for i in checks:
            with self.subTest(i=i):
                self.assertAlmostEqual(pc.flow_transition(*i[0]).magnitude, i[1])

    def test_flow_transition_range(self):
        """flow_transition should not accept inputs <= 0."""
        checks = ((1, 0), (0, 1))
        for i in checks:
            with self.subTest(i=i):
                self.assertRaises(ValueError, pc.flow_transition, *i)

    def test_flow_transition_units(self):
        """flow_transition should handle units correctly."""
        base = pc.flow_transition(2, 0.4).magnitude
        checks = ([2 * u.m, 0.4 * u.m**2/u.s],
                  [200 * u.cm, 0.4],
                  [2, 4000 * u.cm**2/u.s])
        for i in checks:
            with self.subTest(i=i):
                self.assertAlmostEqual(pc.flow_transition(*i).magnitude, base)

    def test_flow_hagen(self):
        """flow_hagen should return known value for known inputs."""
        checks = (([1, 0.4, 5.21, 0.6], 0.03079864403023667),
                  ([0.05, 0.0006, 0.3, 1.1], 2.7351295806397676e-09))
        for i in checks:
            with self.subTest(i=i):
                self.assertAlmostEqual(pc.flow_hagen(*i[0]).magnitude, i[1])

    def test_flow_hagen_range(self):
        """flow_hagen should raise errors when inputs are out of bounds."""
        failChecks = ((0, 1, 1, 1), (1, -1, 1, 1), (1, 1, 0, 1), (1, 1, 1, 0))
        for i in failChecks:
            with self.subTest(i=i):
                self.assertRaises(ValueError, pc.flow_hagen, *i)
        passChecks = ((1, 1, 1, 1), (1, 0, 1, 1))
        for i in passChecks:
            with self.subTest(i=i):
                pc.flow_hagen(*i)

    def test_flow_hagen_units(self):
        """flow_hagen should handle units properly."""
        base = pc.flow_hagen(0.05, 0.0006, 0.3, 1.1).magnitude
        checks = ([0.05 * u.m, 0.0006 * u.m, 0.3 * u.m, 1.1 * u.m**2/u.s],
                  [5 * u.cm, 0.0006, 0.3, 1.1],
                  [0.05, 0.6 * u.mm, 0.3, 1.1],
                  [0.05, 0.0006, 0.0003 * u.km, 1.1],
                  [0.05, 0.0006, 0.3, 11000 * u.cm**2/u.s])
        for i in checks:
            with self.subTest(i=i):
                self.assertAlmostEqual(pc.flow_hagen(*i).magnitude, base)

    def test_flow_swamee(self):
        """flow_swamee should return known value for known inputs."""
        checks = (([2, 0.04, 3, 0.1, 0.37], 2.9565931732010045),)
        for i in checks:
            with self.subTest(i=i):
                self.assertAlmostEqual(pc.flow_swamee(*i[0]).magnitude, i[1])

    def test_flow_swamee_range(self):
        """flow_swamee should raise errors when inputs are out of bounds."""
        failChecks = ((0, 1, 1, 1, 1), (1, 0, 1, 1, 1), (1, 1, 0, 1, 1),
                      (1, 1, 1, 0, 1), (1, 1, 1, 1, -0.1), (1, 1, 1, 1, 2))
        for i in failChecks:
            with self.subTest(i=i):
                self.assertRaises(ValueError, pc.flow_swamee, *i)
        passChecks = ((1, 1, 1, 1, 1), (1, 1, 1, 1, 0))
        for i in passChecks:
            with self.subTest(i=i):
                pc.flow_swamee(*i)

    def test_flow_swamee_units(self):
        """flow_swamee should handle units correctly."""
        base = pc.flow_swamee(2, 0.04, 3, 0.1, 0.37).magnitude
        checks = ([2 * u.m, 0.04 * u.m, 3 * u.m, 0.1 * u.m**2/u.s, 0.37 * u.m],
                  [200 * u.cm, 0.04, 3, 0.1, 0.37],
                  [2, 40 * u.mm, 3, 0.1, 0.37],
                  [2, 0.04, 0.003 * u.km, 0.1, 0.37],
                  [2, 0.04, 3, 1000 * u.cm**2/u.s, 0.37],
                  [2, 0.04, 3, 0.1, 37 * u.cm])
        for i in checks:
            with self.subTest(i=i):
                self.assertAlmostEqual(pc.flow_swamee(*i).magnitude, base)

    def test_flow_pipemajor(self):
        """flow_pipemajor should return known result for known inputs."""
        checks = (([1, 0.97, 0.5, 0.025, 0.06], 18.677652880272845),
                  ([2, 0.62, 0.5, 0.036, 0.23], 62.457206502701297))
        for i in checks:
            with self.subTest(i=i):
                self.assertAlmostEqual(pc.flow_pipemajor(*i[0]).magnitude, i[1])

    def test_flow_pipemajor_units(self):
        """flow_pipemajor should handle units correctly."""
        base = pc.flow_pipemajor(2, 0.62, 0.5, 0.036, 0.23).magnitude
        checks = ([2 * u.m, 0.62 * u.m, 0.5 * u.m,
                   0.036 * u.m**2/u.s, 0.23 * u.m],
                  [200 * u.cm, 0.62, 0.5, 0.036, 0.23],
                  [2, 620 * u.mm, 0.5, 0.036, 0.23],
                  [2, 0.62, 0.0005 * u.km, 0.036, 0.23],
                  [2, 0.62, 0.5, 360 * u.cm**2/u.s, 0.23],
                  [2, 0.62, 0.5, 0.036, 23 * u.cm])
        for i in checks:
            with self.subTest(i=i):
                self.assertAlmostEqual(pc.flow_pipemajor(*i).magnitude, base)

    def test_flow_pipeminor(self):
        """flow_pipeminor should return known results for known input."""
        self.assertAlmostEqual(pc.flow_pipeminor(1, 0.125, 3).magnitude,
                         0.71000203931611083)

    def test_flow_pipeminor_range(self):
        """flow_pipeminor should raise errors when inputs are out of bounds."""
        failChecks = ((1, -1, 1), (1, 1, 0))
        for i in failChecks:
            with self.subTest(i=i):
                self.assertRaises(ValueError, pc.flow_pipeminor, *i)
        passChecks = ((1, 1, 1), (1, 0, 1))
        for i in passChecks:
            with self.subTest(i=i):
                pc.flow_pipeminor(*i)

    def test_flow_pipeminor_units(self):
        """flow_pipeminor should handle units correctly."""
        base = pc.flow_pipeminor(1, 0.125, 3).magnitude
        checks = ((1 * u.m, 0.125 * u.m, 3 * u.dimensionless),
                  (0.001 * u.km, 0.125, 3),
                  (1, 125 * u.mm, 3))
        for i in checks:
            with self.subTest(i=i):
                self.assertAlmostEqual(pc.flow_pipeminor(*i).magnitude, base)

    def test_flow_pipe(self):
        """flow_pipe should return known value for known inputs."""
        checks = (([0.25, 0.4, 2, 0.58, 0.029, 0], 0.000324207170118938),
                  ([0.25, 0.4, 2, 0.58, 0.029, 0.35], 0.000324206539183988))
        for i in checks:
            with self.subTest(i=i):
                self.assertAlmostEqual(pc.flow_pipe(*i[0]).magnitude, i[1])

    def test_flow_pipe_units(self):
        """flow_pipe should handle units correctly."""
        base = pc.flow_pipe(0.25, 0.4, 2, 0.58, 0.029, 0.35).magnitude
        checks = ([0.25 * u.m, 0.4 * u.m, 2 * u.m,
                   0.58 * u.m**2/u.s, 0.029 * u.m, 0.35],
                  [25 * u.cm, 0.4, 2, 0.58, 0.029, 0.35],
                  [0.25, 400 * u.mm, 2, 0.58, 0.029, 0.35],
                  [0.25, 0.4, 0.002 * u.km, 0.58, 0.029, 0.35],
                  [0.25, 0.4, 2, 580000 * u.mm**2/u.s, 0.029, 0.35],
                  [0.25, 0.4, 2, 0.58, 29 * u.mm, 0.35])
        for i in checks:
            with self.subTest(i=i):
                self.assertAlmostEqual(pc.flow_pipe(*i).magnitude, base)


class DiamFuncsTest(unittest.TestCase):
    """Test the diameter functions."""
    def test_diam_hagen(self):
        """diam_hagen should return known value for known inputs."""
        self.assertAlmostEqual(pc.diam_hagen(0.006, 0.00025, 0.75, 0.0004).magnitude,
                         0.4158799465199102)

    def test_diam_hagen_range(self):
        """diam_hagen should raise errors when inputs are out of bounds."""
        failChecks = ((0, 1, 1, 1), (1, 0, 1, 1), (1, 1, 0, 1), (1, 1, 1, 0))
        for i in failChecks:
            with self.subTest(i=i):
                self.assertRaises(ValueError, pc.diam_hagen, *i)
        pc.diam_hagen(1, 1, 1, 1)

    def test_diam_hagen_units(self):
        """diam_hagen should handle units correctly."""
        base = pc.diam_hagen(0.006, 0.00025, 0.75, 0.0004).magnitude
        checks = ([0.006 * u.m**3/u.s, 0.00025 * u.m,
                   0.75 * u.m, 0.0004 * u.m**2/u.s],
                  [6 * u.L/u.s, 0.00025, 0.75, 0.0004],
                  [0.006, 0.25 * u.mm, 0.75, 0.0004],
                  [0.006, 0.00025, 75 * u.cm, 0.0004],
                  [0.006, 0.00025, 0.75, 4 * u.cm**2/u.s])
        for i in checks:
            with self.subTest(i=i):
                self.assertAlmostEqual(pc.diam_hagen(*i).magnitude, base)

    def test_diam_swamee(self):
        """diam_swamee should return known value for known input."""
        self.assertAlmostEqual(pc.diam_swamee(0.06, 1.2, 7, 0.2, 0.0004).magnitude,
                         0.19286307314945772)

    def test_diam_swamee_range(self):
        """diam_swamee should raise errors if inputs are out of bounds."""
        failChecks = ((0, 1, 1, 1, 1), (1, 0, 1, 1, 1), (1, 1, 0, 1, 1),
                      (1, 1, 1, 0, 1), (1, 1, 1, 1, 2), (1, 1, 1, 1, -1))
        for i in failChecks:
            with self.subTest(i=i):
                self.assertRaises(ValueError, pc.diam_swamee, *i)
        passChecks = ((1, 1, 1, 1, 1), (1, 1, 1, 1, 0))
        for i in passChecks:
            with self.subTest(i=i):
                pc.diam_swamee(*i)

    def test_diam_swamee_units(self):
        """diam_swamee should handle units correctly."""
        base = pc.diam_swamee(0.06, 1.2, 7, 0.2, 0.0004).magnitude
        checks = ([0.06 * u.m**3/u.s, 1.2 * u.m, 7 * u.m,
                   0.2 * u.m**2/u.s, 0.0004 * u.m],
                  [60 * u.L/u.s, 1.2, 7, 0.2, 0.0004],
                  [0.06, 0.0012 * u.km, 7, 0.2, 0.0004],
                  [0.06, 1.2, 700 * u.cm, 0.2, 0.0004],
                  [0.06, 1.2, 7, 2000 * u.cm**2/u.s, 0.0004],
                  [0.06, 1.2, 7, 0.2, 0.4 * u.mm])
        for i in checks:
            with self.subTest(i=i):
                self.assertAlmostEqual(pc.diam_swamee(*i).magnitude, base)

    def test_diam_pipemajor(self):
        """diam_pipemajor should return known value for known inputs."""
        checks = (([0.005, 0.03, 1.6, 0.53, 0.002], 0.8753787620849313),
                  ([1, 2, 0.03, 0.004, 0.005], 0.14865504303291951))
        for i in checks:
            with self.subTest(i=i):
                self.assertAlmostEqual(pc.diam_pipemajor(*i[0]).magnitude, i[1])

    def test_diam_pipemajor_units(self):
        """diam_pipemajor should handle units correctly."""
        base = pc.diam_pipemajor(1, 2, 0.03, 0.004, 0.005).magnitude
        checks = ([1 * u.m**3/u.s, 2 * u.m, 0.03 * u.m,
                   0.004 * u.m**2/u.s, 0.005 * u.m],
                  [1000 * u.L/u.s, 2, 0.03, 0.004, 0.005],
                  [1, 0.002 * u.km, 0.03, 0.004, 0.005],
                  [1, 2, 3 * u.cm, 0.004, 0.005],
                  [1, 2, 0.03, 40 * u.cm**2/u.s, 0.005],
                  [1, 2, 0.03, 0.004, 5 * u.mm])
        for i in checks:
            with self.subTest(i=i):
                self.assertAlmostEqual(pc.diam_pipemajor(*i).magnitude, base)

    def test_diam_pipeminor(self):
        """diam_pipeminor should return known value for known inputs."""
        checks = (([0.008, 0.012, 0.93], 0.14229440061589257),
                  ([0.015, 0.3, 0.472], 0.073547549463488848))
        for i in checks:
            with self.subTest(i=i):
                self.assertAlmostEqual(pc.diam_pipeminor(*i[0]).magnitude, i[1])

    def test_diam_pipeminor_range(self):
        """diam_pipeminor should raise errors when inputs are out of bounds."""
        failChecks = ((0, 1, 1), (1, 0, 1), (1, 1, -1))
        for i in failChecks:
            with self.subTest(i=i):
                self.assertRaises(ValueError, pc.diam_pipeminor, *i)
        passChecks = ((1, 1, 1), (1, 1, 0))
        for i in passChecks:
            with self.subTest(i=i):
                pc.diam_pipeminor(*i)

    def test_diam_pipeminor_units(self):
        """diam_pipeminor should handle units correctly."""
        base = pc.diam_pipeminor(0.008, 0.012, 0.93).magnitude
        checks = ([0.008 * u.m**3/u.s, 0.012 * u.m, 0.93 * u.dimensionless],
                  [8 * u.L/u.s, 0.012, 0.93],
                  [0.008, 1.2 * u.cm, 0.93])
        for i in checks:
            with self.subTest(i=i):
                self.assertAlmostEqual(pc.diam_pipeminor(*i).magnitude, base)

    def test_diam_pipe(self):
        """diam_pipe should return known value for known inputs."""
        checks = (([0.007, 0.04, 0.75, 0.16, 0.0079, 0], 0.5434876490369928),
                  ([0.007, 0.04, 0.75, 0.16, 0.0079, 0.8], 0.5436137491479152))
        for i in checks:
            with self.subTest(i=i):
                self.assertAlmostEqual(pc.diam_pipe(*i[0]).magnitude, i[1])

    def test_diam_pipe_units(self):
        """diam_pipe should handle units correctly."""
        base = pc.diam_pipe(0.007, 0.04, 0.75, 0.16, 0.0079, 0.8).magnitude
        checks = ([0.007 * u.m**3/u.s, 0.04 * u.m, 0.75 * u.m,
                   0.16 * u.m**2/u.s, 0.0079 * u.m, 0.8],
                  [7 * u.L/u.s, 0.04, 0.75, 0.16, 0.0079, 0.8],
                  [0.007, 4 * u.cm, 0.75, 0.16, 0.0079, 0.8],
                  [0.007, 0.04, 75 * u.cm, 0.16, 0.0079, 0.8],
                  [0.007, 0.04, 0.75, 1600 * u.cm**2/u.s, 0.0079, 0.8],
                  [0.007, 0.04, 0.75, 0.16, 7.9 * u.mm, 0.8])
        for i in checks:
            with self.subTest(i=i):
                self.assertAlmostEqual(pc.diam_pipe(*i).magnitude, base)

class WeirFuncsTest(unittest.TestCase):
    """Test the weir functions."""
    def test_width_rect_weir(self):
        """width_rect_weir should return known value for known inputs."""
        self.assertAlmostEqual(pc.width_rect_weir(0.005, 0.2).magnitude,
                         0.030, places=3)

    def test_width_rect_weir_range(self):
        """width_rect_weird should raise errors when inputs are out of bounds."""
        checks = ((0, 1), (1, 0))
        for i in checks:
            with self.subTest(i=i):
                self.assertRaises(ValueError, pc.width_rect_weir, *i)
        pc.width_rect_weir(1, 1)

    def test_width_rect_weir_units(self):
        """width_rect_weir should handle units correctly."""
        base = pc.width_rect_weir(0.005, 0.2).magnitude
        checks = ([0.005 * u.m**3/u.s, 0.2 * u.m],
                  [5 * u.L/u.s, 0.2],
                  [0.005, 20 * u.cm])
        for i in checks:
            with self.subTest(i=i):
                self.assertAlmostEqual(pc.width_rect_weir(*i).magnitude, base)

    def test_headloss_weir(self):
        """headloss_weir should return known value for known inputs."""
        self.assertAlmostEqual(pc.headloss_weir(0.005, 1).magnitude,
                         0.019, places=3)

    def test_headloss_weir_range(self):
        """headloss_weir should raise errors when inputs are out of bounds."""
        checks = ((0, 1), (1, 0))
        for i in checks:
            with self.subTest(i=i):
                self.assertRaises(ValueError, pc.headloss_weir, *i)
        pc.headloss_weir(1,1)

    def test_headloss_weir_units(self):
        """headloss_weir should handle units correctly."""
        base = pc.headloss_weir(0.005, 1).magnitude
        checks = ([0.005 * u.m**3/u.s, 1 * u.m],
                  [5 * u.L/u.s, 1],
                  [0.005, 100 * u.cm])
        for i in checks:
            with self.subTest(i=i):
                self.assertAlmostEqual(pc.headloss_weir(*i).magnitude, base)

    def test_flow_rect_weir(self):
        """flow_rect_weir should return known value for known inputs."""
        self.assertAlmostEqual(pc.flow_rect_weir(2, 1).magnitude, 5.261, places=3)

    def test_flow_rect_weir_range(self):
        """flow_rect_weir should raise errors when inputs are out of bounds."""
        checks = ((0, 1), (1, 0))
        for i in checks:
            with self.subTest(i=i):
                self.assertRaises(ValueError, pc.flow_rect_weir, *i)
        pc.flow_rect_weir(1, 1)

    def test_flow_rect_weir_units(self):
        """flow_rect_weir should handle units correctly."""
        base = pc.flow_rect_weir(2, 1).magnitude
        checks = ([2 * u.m, 1 * u.m],
                  [200 * u.cm, 1],
                  [2, 100 * u.cm])
        for i in checks:
            with self.subTest(i=i):
                self.assertAlmostEqual(pc.flow_rect_weir(*i).magnitude, base)


class MiscPhysFuncsTest(unittest.TestCase):
    """Test the miscellaneous physchem functions."""
    def test_height_water_critical(self):
        """height_water_critical should return known value for known inputs."""
        self.assertAlmostEqual(pc.height_water_critical(0.006, 1.2).magnitude,
                         0.013660704939951886)

    def test_height_water_critical_range(self):
        """height_water_critical should raise errors when inputs are out of bounds."""
        checks = ((0, 1), (1, 0))
        for i in checks:
            with self.subTest(i=i):
                self.assertRaises(ValueError, pc.height_water_critical, *i)
        pc.height_water_critical(1, 1)

    def test_height_water_critical_units(self):
        """height_water_critical should handle units correctly."""
        base = pc.height_water_critical(0.006, 1.2).magnitude
        checks = ([0.006 * u.m**3/u.s, 1.2 * u.m],
                  [6 * u.L/u.s, 1.2],
                  [0.006, 120 * u.cm])
        for i in checks:
            with self.subTest(i=i):
                self.assertAlmostEqual(pc.height_water_critical(*i).magnitude, base)

    def test_vel_horizontal(self):
        """vel_horizontal should return known value for known inputs."""
        self.assertAlmostEqual(pc.vel_horizontal(0.03).magnitude, 0.5424016039799292)

    def test_vel_horizontal_range(self):
        """vel_horizontzal should raise an errors when input is <= 0."""
        self.assertRaises(ValueError, pc.vel_horizontal, 0)

    def test_vel_horizontal_units(self):
        """vel_horizontal should handle units correctly."""
        base = pc.vel_horizontal(0.03).magnitude
        checks = (0.03 * u.m, 3 * u.cm)
        for i in checks:
            with self.subTest(i=i):
                self.assertAlmostEqual(pc.vel_horizontal(i).magnitude, base)

    def test_headloss_kozeny(self):
        """headloss_kozeny should return known value for known input."""
        self.assertAlmostEqual(pc.headloss_kozeny(1, 1.4, 0.5, 0.625, 0.8).magnitude,
                         2.1576362645214617)

    def test_headloss_kozeny_range(self):
        """headloss_kozeny should raise errors when inputs are out of bounds."""
        checks = ((0, 1, 1, 1, 1), (1, 0, 1, 1, 1), (1, 1, 0, 1, 1),
                  (1, 1, 1, 1, 0), (1, 1, 1, -1, 1), (1, 1, 1, 2, 1))
        for i in checks:
            with self.subTest(i=i):
                self.assertRaises(ValueError, pc.headloss_kozeny, *i)
        pc.headloss_kozeny(1, 1, 1, 1, 1)

    def test_headloss_kozeny_units(self):
        """headloss_kozeny should handle units correctly."""
        base = pc.headloss_kozeny(1, 1.4, 0.5, 0.625, 0.8).magnitude
        checks = ([1 * u.m, 1.4 * u.m, 0.5 * u.m/u.s,
                   0.625 * u.m, 0.8 * u.m**2/u.s],
                  [100 * u.cm, 1.4, 0.5, 0.625, 0.8],
                  [1, 0.0014 * u.km, 0.5, 0.625, 0.8],
                  [1, 1.4, 30 * u.m/u.min, 0.625, 0.8],
                  [1, 1.4, 0.5, 625 * u.mm, 0.8],
                  [1, 1.4, 0.5, 0.625, 8000 * u.cm**2/u.s])
        for i in checks:
            with self.subTest(i=i):
                self.assertAlmostEqual(pc.headloss_kozeny(*i).magnitude, base)


if __name__ == "__main__":
    unittest.main()
